######################################################################################################################
#                                                                                                                    #
#                                           Autotools buildstep utilities                                            #
#                                                                                                                    #
######################################################################################################################

setup_autotools_aclocaldir() {
	local localdir=$SYSROOT_LOC/usr/share/aclocal
	if [ ! -r $localdir/README ]; then
		if ! $recipe_is_native; then
			localdir="$NATIVE_SYSROOT_LOC${localdir#$SYSROOT_LOC}"
		fi
		if [ ! -r $localdir/README ]; then
			localdir="$WEAK_SYSROOT_LOC${localdir#$NATIVE_SYSROOT_LOC}"
		fi
	fi
	mkdir -p $localdir
	echo $localdir
}
	
setup_autotools_script_path() {
	# We must be careful with relative path names here, as we may be called from
	# either the source directory or the build directory...
	local path=$AUTOTOOLS_SCRIPT_PATH
	if [ -n "$path" ]; then
		case "$path" in
			srcDir/* | workDir/* | buildDir/*)
				# Use working directory names as variable names
				var=${AUTOTOOLS_SCRIPT_PATH%%/*}
				path=${!var}/${path#*/}
				;;
			\$*)# Use paths prefixed by some SW4L directory variable as-is
				;;
			/*) # Use absolute path names as-is too
				;;
			*)	# Place relative path names in the source directory
				path="$srcDir/$path"
				;;
		esac
	else
		path=$srcDir
	fi
	echo $path
}
	
setup_autotools_auxdir() {
	local auxdir=$AUTOTOOLS_AUXDIR
	if [ -n "$auxdir" ]; then
		case "$auxdir" in
			srcDir/* | workDir/* | buildDir/*)
				# Use working directory names as variable names
				var=${auxdir%%/*}
				auxdir=${!var}/${auxdir#*/}
				;;
			\$*)# Use paths prefixed by some SW4L directory variable as-is
				;;
			/*) # Use absolute path names as-is too
				;;
			*)	# Place relative path names in the AUTOTOOLS_SCRIPT_PATH directory
				# but also search it in srcDir, if not found there
				if [ -r "$AUTOTOOLS_SCRIPT_PATH/$auxdir" ]; then
					AUTOTOOLS_AUXDIR="$AUTOTOOLS_SCRIPT_PATH/$auxdir"
				else
					AUTOTOOLS_AUXDIR="$srcDir/$auxdir"
				fi
				;;
		esac
	else
		if [ -d $AUTOTOOLS_SCRIPT_PATH/build-aux ]; then
			auxdir=$AUTOTOOLS_SCRIPT_PATH/build-aux	
		elif [ -d $srcDir/build-aux ]; then
			auxdir=$srcDir/build-aux
		else
			auxdir=$AUTOTOOLS_SCRIPT_PATH
		fi
	fi
	echo $auxdir
}

build_aclocal_options() {
	local acpaths=
	if [ "X${ACLOCAL_DIRS+set}" = "Xset" ]; then
		for i in ${ACLOCAL_DIRS}; do
			i=$(readlink -m $srcDir/$i)
			mkdir -p $srcDir/$i
			acpaths="$acpaths -I $srcDir/$i"
		done
	else
		for i in $(find $srcDir -ignore_readdir_race -maxdepth 2 -name \*.m4 \
		         | grep -v 'aclocal.m4' | grep -v 'acinclude.m4' | sed -e 's,\(.*/\).*$,\1,' | sort -u); do
			acpaths="$acpaths -I $i"
		done
	fi

	acpaths="$acpaths -I $ACLOCALDIR"
	if $recipe_is_native || $recipe_is_cross; then
		acpaths="$acpaths -I ${ACLOCALDIR/\/native\///weak/}"
	fi
	echo "$acpaths"
}

get_aclocal_command() {
	local aclocal=$(which aclocal)
	if [ -z "$aclocal" ]; then aclocal="aclocal"; fi
	aclocal="$aclocal --system-acdir=${ACLOCALDIR}"	
	local autov=$(automake --version | sed -e '1{s/.* //;s/\.[0-9]\+$//};q')
	local automake_lib
	if $recipe_is_cross; then
		automake_lib=${datadir#$TARGET_SYSROOT_LOC}/aclocal-$autov
	else
		automake_lib=${datadir#$SYSROOT_LOC}/aclocal-$autov
	fi
	#automake --version >&2
	#echo "autov is $autov" >&2
	local weakamlib=${WEAK_SYSROOT_LOC}$automake_lib
	if [ -d $NATIVE_SYSROOT_LOC$automake_lib ]; then
		# this should not be needed as it is the default for our aclocal
		AUTOMAKE_LIB=$NATIVE_SYSROOT_LOC$automake_lib
		aclocal="$aclocal --automake-acdir=$AUTOMAKE_LIB"
	else
		if [ -d $weakamlib ]; then
			# This may be needed when adding the weak sysroot
			AUTOMAKE_LIB=$weakamlib
			aclocal="$aclocal --automake-acdir=$weakamlib"
		fi
	fi
	echo "$aclocal"
}

#
# autopoint and others use dates for choosing to overwrite
# existing files, but these dates may not be correct, so we
# may have to do this both before and after calling autoreconf
# Furthermore, the --force flag may suppress this test...
#
copy_gettext_file() {
	local from="$1"
	local to="$2"
	if copy_from_native_sysroot "usr/share/gettext/$from" "$to"; then
		log_exec touch "$to"
	fi
}

do_gettextize() {
	local CONFIGURE_AC=$1
	local PRUNE_M4
	if grep -q "^[[:space:]]*AM_GLIB_GNU_GETTEXT" $CONFIGURE_AC; then
		if grep -q "sed.*POTFILES" $CONFIGURE_AC; then
			: # do nothing -- we still have an old unmodified configure.ac
		else
			log glib-gettextize --force --copy
			echo "no" | glib-gettextize --force --copy
		fi
	elif [ "$recipe_base_name" != "gettext" ] \
	&&   grep -q "^[[:space:]]*AM_GNU_GETTEXT" $CONFIGURE_AC; then
		# We'd call gettextize here if it wasn't so broken...
		local gettextdata=$(search_in_native_sysroot usr/share/gettext)
		if [ -n "$gettextdata" -a -d "$gettextdata" ]; then 
			# This should only fail when building gettext-minimal-native itself
			copy_gettext_file config.rpath ${AUTOTOOLS_AUXDIR}/config.rpath
			if [ -d po ]; then
				copy_gettext_file po/Makefile.in.in po/Makefile.in.in
				if [ ! -e po/remove-potcdate.sin ]; then
					copy_gettext_file po/remove-potcdate.sin po/remove-potcdate.sin
				fi
			fi
			if [ -d runtime-po ]; then # At least bison has such a runtime PO directory
				copy_gettext_file po/Makefile.in.in runtime-po/Makefile.in.in
				if [ ! -e runtime-po/remove-potcdate.sin ]; then
					copy_gettext_file po/remove-potcdate.sin runtime-po/remove-potcdate.sin
				fi
			fi
			PRUNE_M4="$PRUNE_M4 gettext.m4 iconv.m4 lib-ld.m4 lib-link.m4 lib-prefix.m4 nls.m4 po.m4 progtest.m4"
		fi
	fi
	if grep -q "^[[:space:]]*[AI][CT]_PROG_INTLTOOL" $CONFIGURE_AC; then
		if [[ ! " $recipe_buildtime_dependencies " =~ " intltool-native " ]]; then
			warning "Package '$recipe_name' needs 'intltool-native', but dependency is missing"
		fi
		PRUNE_M4="$PRUNE_M4 intltool.m4"
		local intltoolize="$(which intltoolize)"
		if [ -z "$intltoolize" ]; then intltoolize="intltoolize"; fi
		log_exec $intltoolize --copy --force --automake
	fi
	for i in $PRUNE_M4; do
		find . -ignore_readdir_race -name $i -delete
	done
}

#
# autopoint and others use dates for choosing to overwrite
# existing files, but these dates may not be correct, so we
# may have to do this both before and after calling autoreconf
# Furthermore, the --force flag may suppress this test...
#
copy_gettext_file() {
	local from="$1"
	local to="$2"
	if copy_from_native_sysroot "usr/share/gettext/$from" "$to"; then
		log_exec touch "$to"
	fi
}

do_gettextize() {
	local CONFIGURE_AC=$1
	local PRUNE_M4
	if grep -q "^[[:space:]]*AM_GLIB_GNU_GETTEXT" $CONFIGURE_AC; then
		if grep -q "sed.*POTFILES" $CONFIGURE_AC; then
			: # do nothing -- we still have an old unmodified configure.ac
		else
			log glib-gettextize --force --copy
			echo "no" | glib-gettextize --force --copy
		fi
	elif [ "$recipe_base_name" != "gettext" ] \
	&&   grep -q "^[[:space:]]*AM_GNU_GETTEXT" $CONFIGURE_AC; then
		# We'd call gettextize here if it wasn't so broken...
		local gettextdata=$(search_in_native_sysroot usr/share/gettext)
		if [ -n "$gettextdata" -a -d "$gettextdata" ]; then 
			# This should only fail when building gettext-minimal-native itself
			copy_gettext_file config.rpath ${AUTOTOOLS_AUXDIR}/config.rpath
			if [ -d po ]; then
				copy_gettext_file po/Makefile.in.in po/Makefile.in.in
				if [ ! -e po/remove-potcdate.sin ]; then
					copy_gettext_file po/remove-potcdate.sin po/remove-potcdate.sin
				fi
			fi
			if [ -d runtime-po ]; then # At least bison has such a runtime PO directory
				copy_gettext_file po/Makefile.in.in runtime-po/Makefile.in.in
				if [ ! -e runtime-po/remove-potcdate.sin ]; then
					copy_gettext_file po/remove-potcdate.sin runtime-po/remove-potcdate.sin
				fi
			fi
			PRUNE_M4="$PRUNE_M4 gettext.m4 iconv.m4 lib-ld.m4 lib-link.m4 lib-prefix.m4 nls.m4 po.m4 progtest.m4"
		fi
	fi
	if grep -q "^[[:space:]]*[AI][CT]_PROG_INTLTOOL" $CONFIGURE_AC; then
		if [[ ! " $recipe_buildtime_dependencies " =~ " intltool-native " ]]; then
			warning "Package '$recipe_name' needs 'intltool-native', but dependency is missing"
		fi
		PRUNE_M4="$PRUNE_M4 intltool.m4"
		local intltoolize="$(which intltoolize)"
		if [ -z "$intltoolize" ]; then intltoolize="intltoolize"; fi
		log_exec $intltoolize --copy --force --automake
	fi
	for i in $PRUNE_M4; do
		find . -ignore_readdir_race -name $i -delete
	done
}

get_autom4te_perllibdir() {
	# For gnu-configize and aclocal to work correctly, we may have to provide
	# autom4te_perllibdir and datadir to point in the native sysroot...
	# Having -cross versions of all autotools tools may solve that?...
	if $recipe_is_cross; then
		autom4te_perllibdir="$NATIVE_SYSROOT_LOC${datadir#$TARGET_SYSROOT_LOC}/autoconf"
	elif $recipe_is_native; then
		autom4te_perllibdir="$NATIVE_SYSROOT_LOC${datadir#$SYSROOT_LOC}/autoconf"
	else
		autom4te_perllibdir="$NATIVE_SYSROOT_LOC$datadir/autoconf"
	fi
	if [ ! -d $autom4te_perllibdir/Autom4te ]; then
		autom4te_perllibdir="${autom4te_perllibdir/\/native\///weak/}"
		if [ ! -d $autom4te_perllibdir/Autom4te ]; then
			autom4te_perllibdir=
		fi
	fi
	echo "$autom4te_perllibdir"
}

global_autotools_setup() {
	export ACLOCALDIR=$(setup_autotools_aclocaldir)
	log "ACLOCALDIR='$ACLOCALDIR'"
	
	export AUTOTOOLS_SCRIPT_PATH=$(setup_autotools_script_path)
	log "AUTOTOOLS_SCRIPT_PATH='$AUTOTOOLS_SCRIPT_PATH'"
	
	export AUTOTOOLS_AUXDIR=$(setup_autotools_auxdir)
	log "AUTOTOOLS_AUXDIR='$AUTOTOOLS_AUXDIR'"
	
	export acpaths=$(build_aclocal_options) # This variable may be used by append scripts
	log "acpaths='$acpaths'"
	
	export ACLOCAL="$(get_aclocal_command)"
	log "ACLOCAL='$ACLOCAL'"
	
	export autom4te_perllibdir=$(get_autom4te_perllibdir)
	log "autom4te_perllibdir='$autom4te_perllibdir'"
}

# This may be needed for other tasks, not only auto_reconfigure, configure and make so we better define these in all cases
global_autotools_setup

autotools_do_auto_reconfigure_prepend() {
	install_uptodate_config_scripts
	cleanup_buildtree
}

#
# This function regenerate configure scripts for an autotools package
#
execute_autoreconf() {
	local here=$(pwd)
	log_exec cd $AUTOTOOLS_SCRIPT_PATH
	
	local CONFIGURE_AC
	if [ -e configure.in ]; then
		CONFIGURE_AC=configure.in
	else
		CONFIGURE_AC=configure.ac
	fi
		
	# autoreconf is too shy to overwrite aclocal.m4 if it doesn't look
	# like it was auto-generated.  Work around this by blowing it away
	# by hand, unless the package specifically asked not to, usually as
	# it includes locally defined macros not available anywhere else.
	mkdir -p m4
	if [ "$DO_NOT_ACLOCAL" != "true" -a "$DO_NOT_FORCE_ACLOCAL" != "true" ]; then
		if ! [[ " $autoreconf_other_flags " =~ " --exclude=aclocal " ]]; then
			log_exec rm -f aclocal.m4
		fi
	fi
	
	do_gettextize $CONFIGURE_AC

	local aclocal=$(get_aclocal_command)
	if [ "$aclocal" != "$ACLOCAL" ]; then
		warning "ACLOCAL value has changed: '$ACLOCAL' --> '$aclocal'"
	fi
	
	local aclocal_auto_dirs="$(build_aclocal_options)"
	if [ "$aclocal_auto_dirs" != "$acpaths" ]; then
		warning "acpaths value has changed: '$acpaths' --> '$aclocal_auto_dirs'"
	fi
	
	autoreconf_flags="-Wcross --verbose --install --force $autoreconf_other_flags"
	
#	if [ "$FORCE_AUTORECONF" != "false" ]; then
#		autoreconf_flags="$autoreconf_flags --force"
#	fi

	if [ "$DO_AUTOPOINT" != "true" ]; then
		autoreconf_flags="$autoreconf_flags --exclude=autopoint"
	fi

	if [ "$DO_NOT_ACLOCAL" = "true" ]; then
		autoreconf_flags="$autoreconf_flags --exclude=aclocal"
	fi
	
	local am_perllibdir=$(get_autom4te_perllibdir)
	if [ "$am_perllibdir" != "$autom4te_perllibdir" ]; then
		warning "ACLOCAL value has changed: '$autom4te_perllibdir' --> '$am_perllibdir'"
	fi

	local autoreconf="$(which autoreconf)"
	if [ -z "$autoreconf" ]; then autoreconf="autoreconf"; fi
	log_exec $autoreconf $autoreconf_flags $aclocal_auto_dirs
# ACLOCAL, ACLOCAL_PATH and autom4te_perllibdir are all exported (as well as
# acpaths == aclocal_auto_dirs)...
#	if [ -n "$am_perllibdir" ]; then
#		datadir=$(dirname "$am_perllibdir")
#		log_exec "autom4te_perllibdir='$am_perllibdir'" "ACLOCAL_PATH='$ACLOCAL_PATH'" "ACLOCAL='$aclocal'" autoreconf $autoreconf_flags $aclocal_auto_dirs
#	else
#		log_exec "ACLOCAL_PATH='$ACLOCAL_PATH'" "ACLOCAL='$aclocal'" autoreconf $autoreconf_flags $aclocal_auto_dirs
#	fi
	log_exec cd $here 
}

#
# This is currently also done in the configure step but an autoreconf step was
# defined for this to streamline builds. Most autotools packages are converted
# to include such an auto-reconfigure task and the call before configure will
# be suppressed when all platforms and libraries will have been updated.
#
autotools_do_auto_reconfigure() {
	#
	# Note we are conservative here: we don't overwrite locally modified
	# files, in build-aux for example, meaning autoreconf may fail harmlessly
	# If a package need to be forcifully autoreconfed,  define
	# 	FORCE_AUTORECONF=true
	# If you don't want aclocal.m4 to be rebuilt define:
	#	DO_NOT_ACLOCAL=true
	# If you don't want autoreconf to run at all, just define
	#	DO_NOT_AUTORECONF=true
	# This variable implies DO_NOT_ACLOCAL
	#
	if [ "$DO_NOT_AUTORECONF" != "true" ]; then
		if [ \( ! -r "$AUTOTOOLS_SCRIPT_PATH/configure.ac" \) \
		-a   \( ! -r "$AUTOTOOLS_SCRIPT_PATH/configure.in" \) ]; then
			fatal 1 "No autoconf.ac nor autoconf.in; maybe you should set AUTOTOOLS_SCRIPT_PATH"
		fi
		execute_autoreconf
		status=$?
	else
		status=0
	fi
	return $status
}

# If we do NOT run autoreconf, whatever the reason, just propagate the stamp
# or outstamp of the preceding task (apply-patches) to not force rebuild all
# subsequent steps. We must have a way to know our predecessors here but, in
# the meantime just look at apply-patches 
copy_ancestors_state() {
	local mystate=$(recipe_task_state_file)
	rm -f "$mystate.outstamp"
	return 0
	local ancestor_state=$(recipe_task_state_file apply-patches)
	if [ -n "$ancestor_state" ]; then
		if [ -r "$ancestor_state.outstamp" ]; then
			cp "$ancestor_state.outstamp" "$mystate.outstamp"
		else
			cp "$ancestor_state" "$mystate.outstamp"
		fi
	fi
	return 0
}

autotools_do_auto_reconfigure_append() {
	if [ "$DO_NOT_AUTORECONF" = "true" ]; then
		copy_ancestors_state
		return 0
	fi
}

autotools_do_call_configure_prepend() {
	cleanup_buildtree
}

autotools_call_configure() {
	local configure_script
	# Sometime configure use the prefixed version of libtool,
	# but only creates the non-prefixed one... Note that links
	# are created first here for the initial configure call,
	# even if libtool is not used, but they will be normalized
	# later, deleting dead links if there are any.
	if [ -r $srcDir/configure ]; then
		configure_script=$srcDir/configure
	elif [ -r $AUTOTOOLS_SCRIPT_PATH/configure ]; then
		configure_script=$AUTOTOOLS_SCRIPT_PATH/configure
	else
		fatal 1 "No configure script found; maybe you should set AUTOTOOLS_SCRIPT_PATH"
	fi
	if [ ! -r "${configure_host_alias}-libtool" ]; then
		log_exec ln -nfs "libtool" "${configure_host_alias}-libtool"
	fi

	# Choose appropriate default configuration files for autotooled recipes
	declare -a possible_configuration_files
	local c_base c_sub c_variant=
	if $recipe_is_native; then
		c_variant=native
	else
		case $cpu_config_arch in
			arm)
				case $cpu_config_endian in
					big)	c_variant=armeb
							;;
					little)	c_variant=armel
							;;
				esac
				;;
			mips)
				case $cpu_config_endian in
					big)	c_variant=mipseb
							;;
					little)	c_variant=mipsel
							;;
				esac
				;;
		esac
	fi
	for c_base in $recipe_target_arch $cpu_config_arch $c_variant $arch_name endian common; do
		for c_sub in $cpu_config_bits $cpu_config_endian $cpu_config_os $recipe_target_os $cpu_config_libc common; do
			possible_configuration_files+=( $c_base-$c_sub )
		done
		possible_configuration_files+=( $c_base )
	done	
	local config_files config_file config_path config_names
	echo "# Combined SITE_CONFIG file (needed for some old configure scripts)" > $buildDir/.autoconf.config_site
	config_files=
	for config_file in "${possible_configuration_files[@]}"; do # These are optional
		if [[ " $config_names " =~ " $config_file " ]]; then continue; fi
		config_names="$config_names $config_file"
		config_path=$(search_script config/$config_file)
		if [ -n "$config_path" -a -r "$config_path" ]; then
			log "Found default config file '$config_file' coming from '${config_path%/$config_file}'"
			echo ""
			echo "# Combine fragment '$config_file' coming from '${config_path%/$config_file}'"
			echo ""
			cat "$config_path"
			config_files="$buildDir/.autoconf.config_site"
		fi
	done >> $buildDir/.autoconf.config_site
	for config_file in $site_configuration_files; do # These are mandatory
		if [[ " $config_names " =~ " $config_file " ]]; then continue; fi
		config_names="$config_names $config_file"
		config_path=$(search_script config/$config_file)
		if [ -n "$config_path" -a -r "$config_path" ]; then
			log "Found default config file '$config_file' coming from '${config_path%/$config_file}'"
			echo ""
			echo "# Combined fragment '$config_file' coming from '${config_path%/$config_file}'"
			echo ""
			cat "$config_path"
			config_files="$buildDir/.autoconf.config_site"
		else
			fatal 98 "Can't find site autoconf configuration file '$config_file'"
		fi
	done >> $buildDir/.autoconf.config_site
	log "export CONFIG_SITE='$config_files'"
	export CONFIG_SITE="$config_files"
	
	local config_var
	declare -a config_vars
	mapfile -t config_vars < <(split_in_words "$configure_cached_vars")
	for config_var in "${config_vars[@]}"; do
		if [ -z "$config_var" ]; then continue; fi
		log "export $config_var"
		eval export $config_var
	done

	log "export ACLOCAL_PATH='${ACLOCAL_PATH}'"
	export ACLOCAL_PATH
	
	local aclocal=$(get_aclocal_command)
	local aclocal_auto_dirs=$(build_aclocal_options)
	log "export ACLOCAL='$aclocal $aclocal_auto_dirs'" 
	export ACLOCAL="$aclocal $aclocal_auto_dirs" 
	
	# Be careful, configure may choke at ANY unexpected message from
	# the compiler, so be sure we are not in verbose mode :-)
	export use_verbose_compiler=false 
	do_call_configure "$configure_script" "$@"
}

# Buildstep defaults
autotools_do_configure() {
	base_do_configure "$@"
}

