######################################################################################################################
#                                                                                                                    #
#                                           Configure buildstep utilities                                            #
#                                                                                                                    #
######################################################################################################################

dot_config_hash() {
	local dot_config=$1
	# Only update if .config file exists
	if [ -n "$dot_config" -a -r "$dot_config" ]; then
		sed -e 's/#.*$//' -e '/^$/d' "$dot_config" | sort | uniq | sha1sum | sed -e 's/ .*$//'
	else
		# Send static value
		echo "No configuration file"
	fi
}

update_configure_status() {
	local old_config="$1"
	local new_config="$2"
	local dot_config=${3:-.config}
	local configure_task=${recipe_configure_task:-configure}
	
	# Only update if config file exists
	if [ -r "$dot_config" ]; then
		if [ -z "$new_config" ]; then
			new_config="$(dot_config_hash $dot_config)"
		fi
		generate_other_task_output_state $configure_task $new_config
		if [ -n "$old_config" ]; then
			if [ "$new_config" != "$old_config" ]; then
				log "Configuration CHANGED" 
			else
				log "Configuration unchanged"
			fi
		fi
	fi
}

install_uptodate_config_scripts() {
	# WARNING: gross hack follows:
	# An autotools built package, but also some non-autotools packages that provide
	# a manually generated configure script, needs these scripts, however only
	# automake or libtoolize actually install the current versions of them.
	# This is a problem in builds that do not use libtool or automake, in the case
	# where we -need- the latest version of these scripts.  e.g. running a build
	# for a package whose autotools are old, on an x86_64 machine, which the old
	# config.sub does not support.  Work around this by installing them manually
	# regardless.
	for what in config.sub config.guess; do
		if [ -r $NATIVE_SYSROOT_LOC/usr/share/gnu-config/$what ]; then
			for cfg in $(find $srcDir -name $what); do
				rm -f $cfg
				log_exec cp $NATIVE_SYSROOT_LOC/usr/share/gnu-config/$what $cfg
				chmod a+x $cfg
			done
		elif [ -r $WEAK_SYSROOT_LOC/usr/share/gnu-config/$what ]; then
			for cfg in $(find $srcDir -name $what); do
				rm -f $cfg
				log_exec cp $WEAK_SYSROOT_LOC/usr/share/gnu-config/$what $cfg
				chmod a+x $cfg
			done
		fi
	done
}

cleanup_buildtree() {
	local dotfile
	if ! [ "$buildDir" -ef "$srcDir" ]; then
		log "Cleaning the out-of-source build tree '$buildDir'"
		rm -rf $buildDir/*
		for dotfile in $buildDir/.??*; do
			if [ ! -r $dotfile ]; then continue; fi
			if [ "$(basename $dotfile)" = ".ac6_build" ]; then continue; fi
			rm -rf $dotfile
		done
	else
		# At least remove the .la files since automake won't automatically
		# regenerate them even if CFLAGS/LDFLAGS are different
		if [ "${make_clean_broken}" != "true" \
		-a \( -e Makefile -o -e makefile -o -e GNUmakefile \) \
		-a -r .ac6-configure-done-marker ]; then
			make clean
		fi
		find $buildDir -ignore_readdir_race -name \*.la -delete
	fi
}

configured_buildDirs=
do_call_configure() {
	local configure_script="$1"
	shift
	local cfgdir=$(get_relative_path $(dirname "$configure_script"))
	# Some configure scripts seems to incorrectly determine bash path...
	log_run CONFIG_SHELL=/bin/bash $cfgdir/$(basename "$configure_script") "${configure_flags[@]}" "$@"
	status=$?
	if [ $status -ne 0 ]; then
		message "The following config.log files may provide further information."
		message $(find $buildDir -ignore_readdir_race -type f -name config.log)
		fatal $status "configure failed"
	fi
	configured_buildDirs="$configured_buildDirs $(/bin/pwd)"
	touch .ac6-configure-done-marker
}

# Here we test to know if we have an autotooled recipe so that the same script
# can be used whatever the class of the recipe
base_call_configure() {
	local configure_cmd=${configure_command-configure}
	# Note that configure_command may be a command with initial arguments
	local script_name=${configure_cmd%% *}
	if [ -x $srcDir/$script_name ]; then
		do_call_configure "$srcDir/$configure_cmd" "$@"
	else
		log "Nothing to do to configure $package_name"
	fi
}

call_configure() {
	call_buildsubstep "call_configure" "$@"
}

base_do_configure() {
	local here=$(pwd)
	log_exec cd $buildDir
	call_configure "$@"
	log_exec cd $here
}

normalize_local_libtool() {
	local buildDir=$1
	if [ -z "$buildDir" -o ! -d "$buildDir" ]; then return 0; fi
	local libtool=$buildDir/$configure_host_alias-libtool
	
	if [ -r $libtool ]; then
		if [ -r $buildDir/libtool ]; then
			if [ -L $buildDir/libtool ]; then
				log_exec rm -f $buildDir/libtool
			elif [ -L $libtool ]; then
				log_exec rm -f $libtool
				log_exec mv $buildDir/libtool $libtool
			else
				# We have two versions of libtool, probably due to the link we have created earlier
				# and that was needed for some broken configure scripts; cleanup the oldest one
				if [ $buildDir/libtool -nt $libtool ]; then
					log_exec rm -f $libtool
					log_exec mv $buildDir/libtool $libtool
				else
					log_exec rm -f $buildDir/libtool
				fi
			fi
		fi
	elif [ -r $buildDir/libtool ]; then
		log_exec rm -f $libtool
		log_exec mv $buildDir/libtool $libtool
	else
		log_exec rm -f $libtool $buildDir/libtool
		return 0
	fi
	
	# Now ensure that the correct lt_sysroot is set, whatever the --with-libtool-sysroot
	# option passed to the configure script (we keep it in recipe_utils.sh when
	# computing the configure options as the configure script itself may need it...) 
	# Note this is only useful in the (rare) cases where the configure script does not++
	# set the variable correctly, thet is packages that have an old configure
	# script that do not understand the --with-libtool-sysroot flag.
	log_exec sed -e 's!^lt_sysroot=.*$!lt_sysroot='"$SYSROOT_LOC"'!' -i $libtool

	# Now create aliases to the generated libtool, with standard variations
	# as some autotooled packages do not keep the full host_alias...
	libtool=$(basename $libtool)
	local xlibtool=${configure_host_alias%-gnu*}-libtool
	local ylibtool=${xlibtool/-pc-/-}
	local zlibtool=${configure_host_alias/-pc-/-}-libtool
	local _libtool
	for _libtool in $xlibtool $ylibtool $zlibtool libtool; do
		if [ $libtool != $_libtool -a ! -r $buildDir/$_libtool ]; then
			log_exec ln -nfs $libtool $buildDir/$_libtool 
		fi
	done
}

base_do_configure_prepend() {
	local dotfile
	if ! $recipe_is_autotooled; then
		cleanup_buildtree
	fi
	rm -f .ac6-configure-done-marker
	install_uptodate_config_scripts
}

base_do_configure_append() {
	local buildDir
	for buildDir in $configured_buildDirs; do
		normalize_local_libtool $buildDir
	
		#
		# Some Makefiles seems not to create directories in the build tree when needed
		# expecting they already exist as they don't really expect to be build out-of-source-tree
		#
		if ! [ "$buildDir" -ef "$srcDir" ]; then
			local dir
			for dir in $(cd $srcDir; find . -type d | grep -v _build | grep -v _install); do
				[ -d "$dir" ] || mkdir -p "$buildDir/$dir"
			done
		fi
	done
}

# Yocto-compatibility layer
do_configure() {
	do_buildstep "$@"
}
oe_runconf() {
	do_configure "$@"
}
