#####################################################################################################################
#                                                                                                                   #
#                                            Install buildstep utilities                                            #
#                                                                                                                   #
#####################################################################################################################

# Native or cross packages use the sysroot ($(NATIVE|TARGET)_SYSROOT_LOC)
# as prefix, so the install will have created in $installDir subdirectories
# with this name which will contain the install result.
# Here we must suppress this subdirectory so that we just have the installed
# files directly in $installDir, and this only for target packages: usually
# native and cross packages may generate files located in the target sysroot
# but, as these files may be host-dependent, they will not be packaged as the
# correct files will be provided by the equivalent target package. The notable
# exception to this rule is the compiler itself, but it is installed by a
# specific install script and do not go through this function.
normalize_install_tree() {
	rm -rf "$nativeInstallDir" 
	rm -rf "$workDir/_prefix" "$workDir/_build_marks" "$workDir/_removed"
	if [ -n "$NATIVE_SYSROOT_LOC" -a -d "$installDir$NATIVE_SYSROOT_LOC" ] \
	|| [ -n "$TARGET_SYSROOT_LOC" -a -d "$installDir$TARGET_SYSROOT_LOC" ]; then
		# We have a native or cross package, suppress the subdirectory trees
		# and place what should go on the target sysroot in $installDir, while
		# what must be placed in the native sysroot will go in _sysroot
		#
		# Note that, for a cross package, the part targeting the target is
		# in fact compiled for the host so will never be packaged; the same
		# package, compiled for the target, will provide these files

		# First move build auxiliary files out of the way, if they already exist
		# (.fakeroot and .ac6_build may only be created at the end of the first run)
		mkdir -p "$workDir/_build_marks"
		[ -d "$installDir/.ac6_build" ] && mv "$installDir/.ac6_build" "$workDir/_build_marks"
		[ -r "$installDir/.fakeroot" ] && mv "$installDir/.fakeroot" "$workDir/_build_marks"
		
		if [ -n "$NATIVE_SYSROOT_LOC" -a -d "$installDir$NATIVE_SYSROOT_LOC" ]; then
			log_exec mv "$installDir$NATIVE_SYSROOT_LOC" "$nativeInstallDir"
		fi
		if [ -n "$TARGET_SYSROOT_LOC" -a -d "$installDir$TARGET_SYSROOT_LOC" ]; then
			log_exec mv "$installDir$TARGET_SYSROOT_LOC" "$workDir/_prefix"
		fi
		
		# Now check that $installDir only contains empty directories
		if is_tree_empty "$installDir"; then
			rm -rf "$installDir"
			if $recipe_is_cross; then
				if [ -d "$workDir/_prefix" ]; then
					log_exec mv "$workDir/_prefix" "$workDir/_removed"
				fi
				mkdir -p "$installDir"
			elif [ -d "$workDir/_prefix" ]; then
				log_exec mv "$workDir/_prefix" "$installDir"
			else
				mkdir -p "$installDir"
			fi
			# And reinstall saved auxiliary files
			[ -d "$workDir/_build_marks/.ac6_build" ] && mv "$workDir/_build_marks/.ac6_build" "$installDir"
			[ -r "$workDir/_build_marks/.fakeroot" ] && mv "$workDir/_build_marks/.fakeroot" "$installDir"
			rm -rf "$workDir/_build_marks"
		else
			log_error "Some files were installed outside the provided prefixes (lists trimmed to 20 entries)"
			find "$installDir" -type f -ls | sed -e 's|^'"$installDir"'||' | head -n 20 >&2
			log_error "Files moved from target sysroot were:"
			find "$workDir/_prefix" -type f -ls | sed -e 's|^'"$workDir/_prefix"'||' | head -n 20 >&2
			log_error "Files moved from native sysroot were:"
			find "$nativeInstallDir" -type f -ls | sed -e 's|^'"$nativeInstallDir"'||' | head -n 20 >&2
			rm -rf "$workDir/_prefix" "$nativeInstallDir" "$installDir"
			# Reinstall saved auxiliary files
			mkdir -p "$installDir"
			[ -d "$workDir/_build_marks/.ac6_build" ] && mv "$workDir/_build_marks/.ac6_build" "$installDir"
			[ -r "$workDir/_build_marks/.fakeroot" ] && mv "$workDir/_build_marks/.fakeroot" "$installDir"
			rm -rf "$workDir/_build_marks"
			error 2 "Can't install with incorrectly located files"
		fi
	elif $recipe_is_native; then
		# This is a non-autotooled native-only package, installed at the root of _install,
		# place everything in _sysroot except the auxiliary files
		log_exec mv $installDir $nativeInstallDir
		mkdir -p $installDir
		[ -d "$nativeInstallDir/.ac6_build" ] && mv "$nativeInstallDir/.ac6_build" "$installDir"
		[ -r "$nativeInstallDir/.fakeroot" ] && mv "$nativeInstallDir/.fakeroot" "$installDir"
	fi
	
	# Now ensures the files in the install directories are dated now, so that
	# utilities that use dates to know if they must update files will work as
	# expected
	if [ -d $installDir ]; then
		log "find $installDir -type f | xargs -r touch"
		find $installDir -type f | xargs -r touch
	fi
	if [ -d $nativeInstallDir ]; then
		log "find $nativeInstallDir -type f | xargs -r touch"
		find $nativeInstallDir -type f | xargs -r touch
	fi

	# Be careful, if the last test fails, the return status will indicate an error...
	return 0
}

do_install_prepend() {
	# Clean the install location but keep the build markers (so don't remove $installDir)
	log_exec rm -rf $installDir/* $installDir/.fakeroot
}

# This will be executed as the real last step of the install task
do_install_append() {
	# Info dir listing isn't interesting at this point so remove it if it exists.
	if [ -e "$installDir$infodir/dir" ]; then
		log_exec rm -f $installDir$infodir/dir
	fi
	normalize_install_tree
	# Suppress libtool pseudo-libraries that are broken in our environment
	# Now we keep them and patch them when installing in the sysroots or packaging
	#log_exec rm -f $(find $installDir -name '*.la')
}

#
# Install shared library file and create the necessary links
#		do_install_sofile sofile... [ installdir ]
# if omitted, installdir defaults to $libdir
#
# Example: do_install_sofile libfoo.so.1.2.3 $libdir
#
do_install_sofile() {
	declare -a sofiles=("$@")
	local nb_sofiles=${#sofiles[@]]}
	local outdir=$libdir
	if [ $nb_sofiles -gt 1 ]; then
		if [ -d "${sofiles[nb_sofiles-1]}" ]; then
			outdir=${sofiles[--nb_sofiles]}
			unset sofiles[nb_sofiles]
		fi
	fi
	
	local sofile
	for sofile in "${sofiles[@]}"; do
		local libname=$(basename $sofile)
		local libvers=$(echo $libname | sed -n -e 's/^lib.*-\([0-9a-zA-Z][.0-9a-zA-Z-]*\)\.so/\1/p' -e 's/^lib.*\.so[.-]]\([0-9a-zA-Z][.0-9a-zA-Z-]*\)$/\1/p')
		if [ -z "$libvers" ]; then
			error "do_install_sofile: Shared library '$libname' must haved versioned filename (e.g. libfoo.so.1.2.3, libfoo.so-1.2.3 or libfoo-1.2.3.so)"
		fi
		log_exec install -m 755 $sofile $outdir/$libname
		local solink=${libname/$libvers/}
		local sonamelink=$(call_readelf -d $sofile | grep 'Library soname:' |sed -e 's/.*\[\(.*\)\].*/\1/')
		if [ -z $sonamelink ]; then
			error "do_install_sofile: $libname is missing ELF tag 'SONAME'."
		fi
		ln -sf $libname $outdir/$sonamelink
		ln -sf $libname $outdir/$solink
	done
}

#
# Install a library, in all its forms
# Example
#
# do_install_library libltdl $installDir$libdir
# do_install_library -C src/libblah libblah $installDir$libdir
#
do_install_library() {
	local dir libtool silent require_static require_shared
	while [ "$#" -gt 0 ]; do
		case "$1" in
			-C)
				shift
				dir="$1"
				;;
			-s)
				silent=1
				;;
			-a)
				require_static=1
				;;
			-so)
				require_shared=1
				;;
			-*)
				fatal 1 "do_install_library: unknown option: $1"
				;;
			*)
				break;
				;;
		esac
		shift
	done

	libname="$1"
	shift
	destpath="$1"
	if [ -z "$destpath" ]; then
		fatal 2 "do_install_library: no destination path specified"
	fi

	if [ -z "$dir" ]; then
		dir=`pwd`
	fi

	dotlai=$libname.lai

	# Sanity check that the libname.lai is unique
	number_of_files=`(cd $dir; find . -name "$dotlai") | wc -l`
	if [ $number_of_files -gt 1 ]; then
		fatal 3 "do_install_library: $dotlai is not unique in $dir"
	fi


	dir=$dir`(cd $dir;find . -name "$dotlai") | sed "s/^\.//;s/\/$dotlai\$//;q"`
	olddir=`pwd`
	log_exec cd $dir

	lafile=$libname.la

	# If such file doesn't exist, try to cut version suffix
	if [ ! -f "$lafile" ]; then
		libname1=`echo "$libname" | sed 's/-[0-9.]*\$//'`
		lafile1=$libname.la
		if [ -f "$lafile1" ]; then
			libname=$libname1
			lafile=$lafile1
		fi
	fi

	if [ -f "$lafile" ]; then
		# libtool archive
		eval `cat $lafile|grep "^library_names="`
		libtool=1
	else
		library_names="$libname.so* $libname.dll.a $libname.*.dylib"
	fi

	log_exec install -d $destpath/
	dota=$libname.a
	if [ -f "$dota" -o -n "$require_static" ]; then
		rm -f $destpath/$dota
		log_exec install -m 0644 $dota $destpath/
	fi
	if [ -f "$dotlai" -a -n "$libtool" ]; then
		rm -f $destpath/$libname.la
		log_exec install -m 0644 $dotlai $destpath/$libname.la
	fi

	for name in $library_names; do
		files=`eval echo $name`
		for f in $files; do
			if [ ! -e "$f" ]; then
				if [ -n "$libtool" ]; then
					fatal 4 "do_install_library: $dir/$f not found."
				fi
			elif [ -L "$f" ]; then
				log_exec cp -P "$f" $destpath/
			elif [ ! -L "$f" ]; then
				libfile="$f"
				rm -f $destpath/$libfile
				log_exec install -m 0755 $libfile $destpath/
			fi
		done
	done

	if [ -z "$libfile" ]; then
		if  [ -n "$require_shared" ]; then
			fatal 5 "do_install_library: unable to locate shared library"
		fi
	elif [ -z "$libtool" ]; then
		# special case hack for non-libtool .so.#.#.# links
		baselibfile=`basename "$libfile"`
		if (echo $baselibfile | grep -qE '^lib.*\.so\.[0-9.]*\$'); then
			sonamelink=`readelf -d $libfile |grep 'Library soname:' |sed -e 's/.*\[\(.*\)\].*/\1/'`
			solink=`echo $baselibfile | sed -e 's/\.so\..*/.so/'`
			if [ -n "$sonamelink" -a x"$baselibfile" != x"$sonamelink" ]; then
				log_exec ln -sf $baselibfile $destpath/$sonamelink
			fi
			log_exec ln -sf $baselibfile $destpath/$solink
		fi
	fi

	log_exec cd "$olddir"
}

# Yocto-compatibility layer
oe_soinstall() {
	do_install_sofile "$@"
}
oe_libinstall() {
	do_install_library "$@"
}
