# 
# Mason class functions
#

require_script python3_utils.sh

export STAGING_INCDIR=$includedir
export STAGING_LIBDIR=$libdir

export CONFIGURE_FILES="meson.build"
if $recipe_is_cross; then
	MESON_MACHINE=$buildDir/meson.cross
	# Not sure if I must use --cross or --native here...
	MACHINE_FILE="--cross-file ${MESON_MACHINE}"
elif $recipe_is_native; then
	MESON_MACHINE=$buildDir/meson.native
	MACHINE_FILE="--native-file ${MESON_MACHINE}"
else
	MESON_MACHINE=$buildDir/meson.target
	MACHINE_FILE="--cross-file ${MESON_MACHINE}"
fi

# Where the meson.build build configuration is
export MESON_SOURCEPATH="${srcDir}"

# Note buildtype may be:
#	plain			Then you must provide proper optimization, debug, etc. flags
#	debug			Compile with -g -O0
#	debugoptimized	Compile with -g -O2
#	release			Compile with full optimization, stripping exes, ...
MESONOPTS=(
	--buildtype debugoptimized
	--prefix ${prefix}
	--bindir ${bindir#${prefix}/}
	--datadir ${datadir#${prefix}/}
	--includedir ${includedir#${prefix}/}
	--infodir ${infodir#${prefix}/}
	--libdir ${libdir#${prefix}/}
	--libexecdir ${libexecdir#${prefix}/}
	--localedir ${localedir#${prefix}/}
	--localstatedir ${localstatedir#${prefix}/}
	--mandir ${mandir#${prefix}/}
	--sbindir ${sbindir#${prefix}/}
	--sharedstatedir ${sharedstatedir#${prefix}/}
	--sysconfdir ${sysconfdir#${prefix}/}
	--errorlogs
	"-Dc_args=${BUILD_CPPFLAGS} ${BUILD_CFLAGS}"
	"-Dc_link_args=${BUILD_LDFLAGS}"
	"-Dcpp_args=${BUILD_CPPFLAGS} ${BUILD_CXXFLAGS}"
	"-Dcpp_link_args=${BUILD_LDFLAGS}"
)

export MESON_C_ARGS="${CFLAGS}"
export MESON_CPP_ARGS="${CXXFLAGS}"
export MESON_LINK_ARGS="${LDFLAGS}"

meson_array() {
	local arg args
	for arg; do
		args+=", '$arg'"
	done
	printf "[ %s ]\n" "${args#, }"
}

meson_normalize_arch() {
	local bld_arch="$1"
	case $bld_arch in
		ix86)	bld_arch="x86"	;;
	esac
	echo $bld_arch
}

meson_do_write_config() {
	cat > ${MESON_MACHINE} <<- EOF
		#
		# Automatically generated - Do NOT edit
		#

		[build_machine]
			system = '$bld_config_os'
			cpu_family = '$(meson_normalize_arch $bld_config_arch)'
			cpu = '$recipe_build_arch'
			endian = '$bld_config_endian' 
	EOF
	if $recipe_is_native || $recipe_is_cross; then
	    cat >> ${MESON_MACHINE} <<- EOF
	    
			[binaries]
				c = '${native_build_alias}-gcc'
				cpp = '${native_build_alias}-g++'
				ar = '${native_build_alias}-ar'
				nm = '${native_build_alias}-nm'
				strip = '${native_build_alias}-strip'
				readelf = '${native_build_alias}-readelf'
				pkgconfig = '${native_build_alias}-pkg-config'

			[host_machine]
				system = '$bld_config_os'
				cpu_family = '$(meson_normalize_arch $bld_config_arch)'
				cpu = '$recipe_build_arch'
				endian = '$bld_config_endian' 
		EOF
	else # Normal cross-compiled package
	    cat >> ${MESON_MACHINE} <<- EOF
	    
			[binaries]
				c = '${arch_compiler_prefix}-gcc'
				cpp = '${arch_compiler_prefix}-g++'
				ar = '${arch_compiler_prefix}-ar'
				nm = '${arch_compiler_prefix}-nm'
				strip = '${arch_compiler_prefix}-strip'
				readelf = '${arch_compiler_prefix}-readelf'
				pkgconfig = '${arch_compiler_prefix}-pkg-config'

			[host_machine]
				system = 'linux-gnueabi'
				cpu_family = '$(meson_normalize_arch ${arch_name})'
				cpu = '${arch_name}'
				endian = '${cpu_config_endian}'
		EOF
	fi
	if $recipe_is_cross; then
	    cat >> ${MESON_MACHINE} <<- EOF
	    
			[target_machine]
				system = 'linux-gnueabi'
				cpu_family = '$(meson_normalize_arch ${arch_name})'
				cpu = '${arch_name}'
				endian = '${cpu_config_endian}'
		EOF
	fi

	cat >> ${MESON_MACHINE} <<- EOF

		[properties]
			c_args = $(meson_array $MESON_C_ARGS)
			c_link_args = $(meson_array $MESON_LINK_ARGS)
			cpp_args = $(meson_array $MESON_CPP_ARGS)
			cpp_link_args = $(meson_array $MESON_LINK_ARGS)
	EOF
	if $recipe_is_native || $recipe_is_cross; then
		cat >> ${MESON_MACHINE} <<- EOF
			needs_exe_wrapper = false
		EOF
	else
		cat >> ${MESON_MACHINE} <<- EOF
			needs_exe_wrapper = true
			gtkdoc_exe_wrapper = '${buildDir}/gtkdoc-qemuwrapper'
			exe_wrapper = 'qemu'
			sys_toot = '$SYSROOT_LOC'
		EOF
	fi
}

meson_override_native_tools() {
    # Set these so that meson uses the native tools for its build sanity tests,
    # which require executables to be runnable. The cross file will still
    # override these for the target build.
    export CC="${BUILD_CC}"
    export CXX="${BUILD_CXX}"
    export LD="${BUILD_LD}"
    export AR="${BUILD_AR}"
    # These contain *target* flags but will be used as *native* flags.  The
    # correct native flags will be passed via -Dc_args and so on, unset them so
    # they don't interfere with tools invoked by Meson (such as g-ir-scanner)
    unset CPPFLAGS CFLAGS CXXFLAGS LDFLAGS
}

meson_do_configure() {
    # Meson requires native toolset to configure, temporary unset the following variables
    # to avoid using cross compiler to build native executable (eg. sanity check tool)
    unset AR LD NM CC CPP CXX READELF STRIP
    
	if $recipe_is_native || $recipe_is_cross; then
		meson_override_native_tools
	else
		export PKG_CONFIG="${configure_build_prefix}pkg-config"
	fi
	meson_do_write_config

    # Work around "Meson fails if /tmp is mounted with noexec #2972"
    mkdir -p "${buildDir}/meson-private/tmp"
    export TMPDIR="${buildDir}/meson-private/tmp"
    export STAGING_INCDIR
	export STAGING_LIBDIR
	
    log Executing meson ${meson_extra_flags}...
	log_exec meson setup ${MACHINE_FILE} "${MESONOPTS[@]}" ${meson_extra_flags} "${buildDir}" "${MESON_SOURCEPATH}"
}

meson_do_compile() {
    export STAGING_INCDIR
	export STAGING_LIBDIR
    log_exec ninja -v -j ${PARALLEL_MAKE} -C ${buildDir}
}

meson_do_install() {
	export STAGING_INCDIR
	export STAGING_LIBDIR
    DESTDIR="${installDir}" \
    log_exec ninja -v -j ${PARALLEL_MAKE} -C ${buildDir} install
}