#
# Apply patch $patchFile, located in directory $patchBaseDir to tree at $treeToPatch
#
# Patches located in subdirectories of $patchBaseDir are applied in the corresponding 
# subdirectory of $treeToPatch. As usual patches are generated by diff-ing two subtrees.
#
# Record the patches once applied in $patchStatusDir
#
: ${PATCHES_LOC:=$recipe_metadatadir/patches}

require_script_once "version_utils.sh"
		
# Analyze options - These need to be global variables to be assigned by the pre/post_analyze
# functions thereafter and then used in apply_patches later on...
patchStripLevel="-p1"
patchOptNo=0
declare -g -a patchOpts=() # This script is sourced in a function; without -g the variable would be local

pre_analyze_patch_option() {
	local option="$*"
	case $option in
		striplevel=*)
			patchStripLevel="--strip=${option#*=}"
			return 0
			;;
		nostrip)
			patchStripLevel=
			return 0
			;;
	esac
	return 1
}

post_analyze_patch_option() {
	local option="$*"
	case $option in
		*)
			patchOpts[patchOptNo++]="--$option"
			;;
	esac
}

# We have to backup files with a '.orig' extension as some build-systems (like CMake) may build
# all source files they find in the source directory; however we build in a subdirectory of the
# source directory and place backups in a subdirectory of the build directory... 
apply_patches() {
	local treeToPatch="$1"
	shift
	local patchBaseDir="$1"
	shift
	local patchStatusBase="$1"
	shift
	local patchStatusBaseDir="$patchStatusBase/applied_patches"
	local patchBackupBaseDir="$patchStatusBase/backups"
	local patchFile
	
	for patchFile; do
		patchFile="${patchFile//\/.\//\/}"
		patchFile="${patchFile#./}"
		local parentFolder=$(dirname "$patchFile")
		local patchName=$(basename "$patchFile")
		local targetDir="$treeToPatch/$parentFolder"
		local patchStatusDir="$patchStatusBaseDir/$parentFolder"
		local patchStatus="$patchStatusBaseDir/$patchFile"
		local patchBackup="$patchBackupBaseDir/$patchFile"
		
		# Analyze options
		patchStripLevel="-p1"
		patchOptNo=0
		patchOpts=()
				
		if filter_on_options "$patchName" pre_analyze_patch_option post_analyze_patch_option; then
			log "Ignoring patch '$patchName'"
			continue;
		fi
		
		# Build patch options
		if [ -n "$patchStripLevel" ]; then
			patchOpts[patchOptNo++]="$patchStripLevel"
		fi
		patchOpts[patchOptNo++]="--input=$patchBaseDir/$patchFile"
		patchOpts[patchOptNo++]="--forward"
		patchOpts[patchOptNo++]="--backup"
		patchOpts[patchOptNo++]="--version-control=simple"
		patchOpts[patchOptNo++]="--suffix=.orig"
		patchOpts[patchOptNo++]="--prefix=$patchBackup/"
		patchOpts[patchOptNo++]="--directory=$treeToPatch/$parentFolder"
		
		# Apply patch
		mkdir -p "$treeToPatch/$parentFolder"
		
		# Make backups in a specific subtree to allow reliably reverting patches; 
		# problem is that new files are neither backed-up nor revertible, so we must
		# also keep a list of touched files...
		rm -rf "$patchStatus" "$patchBackup"
		mkdir -p "$patchBackup"
		sed -n -e 's|---[[:space:]][^/]\+/\([^[:space:]]\+\).*$|\1|p' \
		       -e 's|+++[[:space:]][^/]\+/\([^[:space:]]\+\).*$|\1|p' < "$patchBaseDir/$patchFile" \
		| sort | uniq > "$patchBackup.changed_files"
		log patch "${patchOpts[@]}"
		if patch "${patchOpts[@]}" >&2; then
			# Now record the applied patch 
			mkdir -p "$patchStatusDir"
			cp "$patchBaseDir/$patchFile" "$patchStatus"
			touch "$patchStatusBase/applied_list"
			echo "$patchFile" >> "$patchStatusBase/applied_list"
		else
			# Here we should unapply the not-fully-applied patch:
			# First remove modified files
			(cd "$targetDir"; rm -f $(cat "$patchBackup.changed_files"))
			# Then reinstall backupped files
			tar -cf - -C "$patchBackup" --transform='s/\.orig$//' . | tar xf - -C "$targetDir"
			# And clean up patch status
			rm -rf "$patchBackup.changed_files" "$patchBackup"
			return 1
		fi
		echo "Applied patch '$patchFile'"
	done
}

#
# Unapply patches - patches must be unapplied in reverse order
#
unapply_patches() {
	local treeToPatch="$1"
	shift
	local patchStatusBase="$1"
	shift
	local patchStatusBaseDir="$patchStatusBase/applied_patches"
	local patchBackupBaseDir="$patchStatusBase/backups"
	local nbPatches=$#
	local -a allPatches
	local patchFile
	
	if [ $nbPatches -eq 0 ]; then return 0; fi
	
	# First reverse the list of patches to unapply
	for patchFile; do
		allPatches[--nbPatches]="$patchFile"
	done
	
	# Then unapply all patches
	for patchFile in "${allPatches[@]}"; do
		local parentFolder=$(dirname "$patchFile")
		local patchName=$(basename "$patchFile")
		local targetDir="$treeToPatch/$parentFolder"
		local patchStatusDir="$patchStatusBaseDir/$parentFolder"
		local patchStatus="$patchStatusBaseDir/$patchFile"
		local patchBackup="$patchBackupBaseDir/$patchFile"

		if [ -r "$patchBackup.changed_files" -a -d "$patchBackup" ]; then
			# Revert the patch by using backups
			# First remove modified patches
			(cd "$targetDir"; rm -f $(cat "$patchBackup.changed_files"))
			# Then reinstall backupped files
			tar -cf - -C "$patchBackup" --transform='s/\.orig$//' . | tar xf - -C "$targetDir"
		else
			# No available backup, apply patch "reversed"
			
			# First analyze options
			local patchOptionsString="${patchName#*.patch@}"
			local -a patchOptions=() patchOpts=()
			local patchStripLevel="-p1"
			local patchOptNo=0
			if [ "$patchOptionsString" != "$patchName" ]; then
				patchOptions=( ${patchOptionsString/;/ } )
				local option
				for option in "${patchOptions[@]}"; do
					if [ -z "$option" ]; then continue; fi
					case $option in
						striplevel=*)
							patchStripLevel="--strip=${option#*=}"
							;;
						nostrip)
							patchStripLevel=
							;;
						*=* | *\<* | *\>*)
							# Skip condition flags
							;;
						*)
							patchOpts[patchOptNo++]="--$option"
							;;
					esac
				done
			fi
			if [ -n "$patchStripLevel" ]; then
				patchOpts[patchOptNo++]=$patchStripLevel
			fi
			patchOpts[patchOptNo++]="--input=$patchStatus"
			patchOpts[patchOptNo++]="--reverse"
			patchOpts[patchOptNo++]="--remove-empty-files"
			patchOpts[patchOptNo++]="--directory=$treeToPatch/$parentFolder"
			
			log_exec patch "${patchOpts[@]}"
		fi
	
		# Now record the patch was no more present
		rm -rf "$patchStatus" "$patchBackup.changed_files" "$patchBackup"
		echo "Removed patch '$patchFile'"
	done
}

if [ -n "$KEEP_PATCHES" ]; then
	case "$KEEP_PATCHES" in
		false)	KEEP_PATCHES=	;;
		true)	KEEP_PATCHES=.	;;
		*)						;;
	esac
fi

find_patch_file() {
	local patchBaseDir="$1"
	local subDir="$2"
	local baseFile="$3"
	local patch
	patch=$(machine_specific_file $patchBaseDir/$subDir $baseFile)
	if [ -n "$patch" -a -r "$patch" ]; then
		patch="${patch#$patchBaseDir/}"
		patch=${patch#./}
		patch=${patch#./}
		patch=${patch//\/.\//\/}
		patch=${patch//\/.\//\/}
		echo "$patch"
	fi
}

copy_patch() {
	local patchBaseDir="$1"
	local patchFile="$2"
	local baseFile=$(basename $patchFile)
	local subDir=$(dirname $patchFile)
	local patch
	patch=$(find_patch_file $patchBaseDir $subDir $baseFile)
	if [ -z "$patch" ]; then
		fatal 4 "Patch file '$patchFile' not found..."
	fi
	if [ -n "$KEEP_PATCHES" ]; then
		mkdir -p $workDir/$KEEP_PATCHES/$subDir
		cp $patchBaseDir/$patch $workDir/$KEEP_PATCHES/$subDir/${baseFile#p[0-9][0-9][0-9][0-9]-}
	fi
	echo "$patch"
}

copy_disabled_patch() {
	local patchBaseDir="$1"
	local patchFile="$(trim ${2##})"
	local baseFile=$(basename $patchFile)
	local subDir=$(dirname $patchFile)
	local patch
	patch=$(find_patch_file $patchBaseDir $subDir $baseFile)
	if [ -z "$patch" ]; then
		# Disabled patches may, optionally, be in disabled_patches instead of patches
		patchBaseDir="${patchBaseDir#/patches}/disabled_patches"
		patch=$(find_patch_file $patchBaseDir_patches $subDir $baseFile)
		if [ -z "$patch" ]; then
			warning "Disabled patch '$patchFile' not found..."
			return
		fi
	fi
	mkdir -p $workDir/${KEEP_PATCHES:-.}/$subDir
	cp $patchBaseDir/$patch $workDir/${KEEP_PATCHES:-.}/$subDir/${baseFile#p[0-9][0-9][0-9][0-9]-}
}

patch_recipe() {
	local treeToPatch="$(absolute_path $1)"
	local patchBaseDir="$(absolute_path $2)"
	local patchStatusBase="$(dirname $(recipe_task_state_file))"
	local patchStatusDir="$patchStatusBase/applied_patches"
	local patchFile 
	local nbPatches=0 nbApplied=0
	local -a allPatches appliedPatches

	if [ ! -d "$patchBaseDir" ]; then return 0; fi
	if [ "$PATCH_TOOL" = "git" ]; then
		warning "Patching using GIT is not yet supported; reverting to standard patch application"
	elif [ "$PATCH_TOOL" = "quilt" ]; then
		message "Quilt is not yet supported; applying patches using 'patch'"
	fi

	#
	# WARNING: Here we do not support alternate patch selection; this should be done to properly import SoC-specific patches
	# that may be provided by Yoto layers for generic recipes; the same way we should support various "series" files
	# for when importing patches from Yocto in a quilt-compatible way.
	#
	local patch_sequence="$patchBaseDir/patch_sequence"
	if [ ! -r "$patch_sequence" ]; then
		# Note that alternate files only work correctly when an explicit patch_sequence is provided
		# Creating the correct patch_sequence here would need to isolate the target-specific subtree
		# from the source-related subtree but may be ignored as no existing recipe need it and most
		# new recipes will come with a patch-sequence.
		patch_sequence="$tmpdir/patch_sequence"
		(
			local find_opts=(-name '*.patch' -o -name '*.patch@*' -o -name '*.diff' -o -name '*.diff@*')
			for patchFile in $(cd "$patchBaseDir"; find . "${find_opts[@]}" | sed -e 's|^\./||'); do 
				if [ -r "$patchBaseDir/$patchFile" ]; then 
					echo "$patchFile"
				fi
			done
			if [ -d "$patchBaseDir/../disabled_patches" ]; then
				for patchFile in $(cd "$patchBaseDir/../disabled_patches"; find . "${find_opts[@]}" | sed -e 's|^\./||'); do 
					if [ -r "$patchBaseDir/$patchFile" ]; then 
						echo "$patchFile #disabled"
					fi
				done
			fi
		) | sort | sed -e 's/^\(.*\) #disabled/# \1/' > $patch_sequence
		# Note the sort above does not work in complex cases, as it sort per directory, not globally...
		# Again this should not cause any problem, as new recipes should have a patch_sequence file
	fi
	local baseFile
	while read patchFile; do
		if [ "${patchFile##}" != "${patchFile}" ]; then
			# Always copy unapplied patches on $workDir, without the (optional) pxxxx- prefix
			copy_disabled_patch $patchBaseDir/../disabled_patches $patchFile
			continue
		fi
		patch=$(copy_patch $patchBaseDir $patchFile)
		if [ -n "$patch" ]; then
			allPatches[nbPatches++]="$patch"
		fi
	done < "$patch_sequence"
	
	mkdir -p "$patchStatusDir"
#	echo "$patchStatusDir" >&2; (cd "$patchStatusDir"; find . -name '*.patch' -o -name '*.patch@*' | sed -e 's|^\./||' | sort) >&2

	local needCleaning=false
	local applied_list
	if [ -r "$patchStatusBase/applied_list" ]; then
		applied_list="$patchStatusBase/applied_list"
	else
		applied_list="$tmpdir/applied_list"
		touch $applied_list
		# Note the sort here does not work in complex cases, as it sort per directory, not globally; hopefully this
		# is only a temporary measure, soon we'll have an applied_list file in all cases...
		for patchFile in $(cd "$patchStatusDir"; find . -name '*.patch' -o -name '*.patch@*' | sed -e 's|^\./||' | sort); do 
			if [ -r "$patchStatusDir/$patchFile" ]; then
				echo "$patchFile" >> $applied_list
			fi
		done
	fi
	while read patchFile; do 
		if [ -r "$patchStatusDir/$patchFile" ]; then
			if [ "$patchFile" != "${allPatches[nbApplied]}" ]; then
				needCleaning=true
			elif ! cmp -s "$patchStatusDir/$patchFile" "$patchBaseDir/$patchFile"; then
				# The patch to apply was different from the already applied one, so
				# don't consider it as already applied, but note something was done
				# on the source code, so all applied patches will have to be unapplied
				needCleaning=true
			fi
			appliedPatches[nbApplied++]="$patchFile"
		else
			needCleaning=true # Only possible if I have an installed list file
		fi
	done < $applied_list
	if $needCleaning; then
		#
		# The set of patches to be applied is not an exact superset of the set of 
		# already applied patches, so we must suppress all applied patches to 
		# restore the source tree in its original state, then apply all patches
		#
		# It should be possible to be a bit more clever and keeop the common set of 
		# inital patches, unapplying the following already applied patches and applying
		# the needed new patches but this should only occur very rarely so why bother?
		#
		if [ $nbApplied -ne 0 ]; then
			echo "Remove already applied patches"
			unapply_patches "$treeToPatch" "$patchStatusBase" "${appliedPatches[@]}"
			echo
			nbApplied=0
			rm -f $applied_list
			rm -rf "$patchStatusBase/applied_patches" "$patchStatusBase/backups"
			mkdir -p "$patchStatusBase/applied_patches" "$patchStatusBase/backups"
		fi
	else
		if ! [ "$applied_list" -ef "$patchStatusBase/applied_list" ]; then
			cp "$applied_list" "$patchStatusBase/applied_list"
		fi
	fi

	if [ $nbPatches -gt $nbApplied ]; then
		echo "Apply patches"
		apply_patches "$treeToPatch" "$patchBaseDir" "$patchStatusBase" "${allPatches[@]:$nbApplied}"
	fi
}
