#! /bin/bash
# Script to run populate sysroot step
#
# PopulateSysroot starts from the install directory in $installDir, copies it into 
# a staging directory in $PKG_STAGING_LOC and suppress a few useless directories as
# well as all files/directories listed in $UNNEEDED_IN_SYSROOT (these must be absolute
# path names starting from the root of the sysroot).
#
# Then for native packages, if $NO_FIXUP_NEEDED is not defined to true, all scripts in the
# resulting sysroot (that is all files starting by '#! /' or '#!/' as well as all files
# listed in $FILES_NEEDING_FIXUP) are edited so that absolute path names starting by /bin,
# /usr/bin, /etc, /usr/share, or by paths listed in $NATIVE_PATHS_TO_RELOCATE are 
# relocated to reference the native sysroot, while absolute path names starting by /lib,
# /usr/lib or by paths listed in $TARGET_PATHS_TO_RELOCATE are relocated to reference the
# platform target sysroot. Note that matching is greedy, that is the longest matching prefix
# determines to which sysroot the path will be relocated. 
#
# Then the populate_sysroot-prepend.sh script is executed if it exists.
#
# Finally the previously populated files are suppressed and the staging directory is moved
# to the appropriate sysroot and the list of all installed files is placed in 
# $loaded_tarballs/$recipe_id/sysroot.list (to allow removing them before re-populating the sysroot.
#

# Give default values to parameters (:= if null or unset, = if only unset, variables must be defined)
: ${omit_subpackages_from_sysroot:="doc test"}
: ${omit_packages_from_sysroot:=""}

stagingDir=$workDir/_staging.$$
silent_do_locked $installDir rm -rf $stagingDir # Just in case
mkdir -p $stagingDir
record_cleanup "silent_do_locked $installDir rm -rf $stagingDir"

TARGET_STAGING_LOC="$stagingDir/target"
mkdir -p $TARGET_STAGING_LOC

NATIVE_STAGING_LOC="$stagingDir/native"
mkdir -p $NATIVE_STAGING_LOC

if $recipe_is_native; then
	STAGING_LOC="$NATIVE_STAGING_LOC"
else
	STAGING_LOC="$TARGET_STAGING_LOC"
fi

PKG_INSTALL_LOC="$TARGET_SYSROOT_LOC"
require_script_once "packaging_utils.sh" # Needed for set_package_install_directory and wipe_recipes_output
require_script_once "package_utils.sh"   # Needed for adjust_libtool_scripts_and_rpath_for_sysroot

if $recipe_is_dummy; then
	# Ensure everything will happen in the WEAK sysroot (even target parts, that should not exist anyway)
	SYSROOT_LOC=$WEAK_SYSROOT_LOC
	NATIVE_SYSROOT_LOC=$WEAK_SYSROOT_LOC
	TARGET_SYSROOT_LOC=$WEAK_SYSROOT_LOC
fi

log "Populating sysroot '$SYSROOT_LOC'"

create_staging_directory() {
	log "Create sysroot staging directories"
	# Note we must copy even empty directories in this step: if the user created them
	# that must be for a reason so don't try to be too smart...
	
	local sysroot_exclusions="/usr/share/man /usr/share/doc /usr/share/info \
	                          /usr/share/locale /usr/share/applications \
	                          /usr/share/fonts /usr/share/pixmaps"
	if is_tree_not_empty $installDir; then
		copy_tree_excluding -d $installDir $TARGET_STAGING_LOC $sysroot_exclusions $other_sysroot_exclusions
	fi
	if is_tree_not_empty $nativeInstallDir; then
		copy_tree_excluding -d $nativeInstallDir $NATIVE_STAGING_LOC $sysroot_exclusions $other_sysroot_exclusions
	fi
}

#
# Relocate native pcfiles
#
make_pcfiles_relocatable() {
	if $recipe_is_native; then
		local subdir rel
		for subdir in $libdir $datadir; do
			subdir=${subdir#$NATIVE_SYSROOT_LOC}/pkgconfig
			if [ -d ${NATIVE_STAGING_LOC}${subdir} ]; then
				rel=$(get_up_path $subdir)
				rel=${rel#/}
				sed -i -e "s:${NATIVE_SYSROOT_LOC}:\${pcfiledir}/$rel:g" ${NATIVE_STAGING_LOC}${subdir}/*.pc
			fi
		done
	fi
}

do_populate_sysroot_prepend() {
	create_staging_directory
	make_pcfiles_relocatable
}

# The prepend buildstep script will be executed just after creating the staging space,
# so it must work on the staging area

do_populate_sysroot() {
	# If we populate the target sysroot for a standard package, libtool scripts
	# and binaries may have to be adjusted to refer to the sysroot and not
	# the target rootfs
	# If we have a native or cross package with target parts from $INSTALL_LOC
	# to populate in the target sysroot these were usually generated with the 
	# corect RPATHs, but we adjust them anyway, just in case
	if is_tree_not_empty $TARGET_STAGING_LOC; then
		adjust_libtool_scripts_and_rpath_for_sysroot -l $installDir $TARGET_STAGING_LOC $TARGET_SYSROOT_LOC
	fi
	
	# The native parts should not to have to be adjusted (if generated to target
	# the native sysroot), but again this does not hurt to adjust if needed
	# Locking is not really needed here, as we will not package these files, 
	# but it does not hurt being consistent...
	if is_tree_not_empty $NATIVE_STAGING_LOC; then
		adjust_libtool_scripts_and_rpath_for_sysroot -l $installDir $NATIVE_STAGING_LOC $NATIVE_SYSROOT_LOC
	fi
}

# The buildstep append script will be executed before copying the staging space in the sysroot
	
do_populate_sysroot_append() {
	log "Suppress previously populated output from '$recipe_id'"
	
	# Be careful that deleting hard links and creating new ones (by a cpio -pdlm) will change
	# the inode of the files; we must lock the corresponding install directory otherwise we
	# may cause errors in the split task where tar will check the file content is not changed
	# Be careful, the already installed version may be an older version...
	if is_tree_not_empty $TARGET_STAGING_LOC; then
		set_package_install_directory $TARGET_SYSROOT_LOC
		wipe_recipes_output -l "$installDir" "${recipe_id%_*}_*"
	fi
	if is_tree_not_empty $NATIVE_STAGING_LOC; then
		set_package_install_directory $NATIVE_SYSROOT_LOC
		wipe_recipes_output -l "$installDir" "${recipe_id%_*}_*"
	fi

	log "Then install new content"
	if is_tree_not_empty $TARGET_STAGING_LOC; then
		set_package_install_directory $TARGET_SYSROOT_LOC
		# Ensure the standard library directory exist, for the sake of RPATH definition
		mkdir -p $TARGET_SYSROOT_LOC/lib $TARGET_SYSROOT_LOC/usr/lib
		mkdir -p $loaded_tarballs/$recipe_id
		copy_tree_excluding -d -v -l $installDir $TARGET_STAGING_LOC $TARGET_SYSROOT_LOC | grep -v '/$' > $loaded_tarballs/$recipe_id/$recipe_id.sysroot.list
		silent_do_locked $installDir rm -rf $TARGET_STAGING_LOC
	fi
	if is_tree_not_empty $NATIVE_STAGING_LOC; then
		set_package_install_directory $NATIVE_SYSROOT_LOC
		# Ensure the standard library directories exist, for the sake of RPATH definition
		mkdir -p $NATIVE_SYSROOT_LOC/lib $NATIVE_SYSROOT_LOC/usr/lib
		mkdir -p $loaded_tarballs/$recipe_id
		copy_tree_excluding -d -v -l $installDir $NATIVE_STAGING_LOC $NATIVE_SYSROOT_LOC | grep -v '/$' > $loaded_tarballs/$recipe_id/$recipe_id.sysroot.list
		silent_do_locked $installDir rm -rf $NATIVE_STAGING_LOC
	fi
}

