######################################################################################################################
#                                                                                                                    #
#                                                    GIT utilities                                                   #
#                                                                                                                    #
######################################################################################################################

DEFAULT_GIT_MIRRORS="
	https://source.codeaurora.org/external|git://git.ac6.fr/mirrors
	https://source.codeaurora.org/external|git://mirrors.ac6-tools.com
	https://source.codeaurora.org/external|http://mirrors.ac6-tools.com
	https://source.codeaurora.org/external|http://www.ac6-tools.com/mirrors
"

# The name is compatible with the Yocto name to benefit from their mirrors...
git_base_name() {
	local gitUrl="${1:-$srcUrl}"
	local urlSite="${gitUrl#*://}"
	local urlPath="/${urlSite#*/}"
	urlSite="${urlSite%%/*}"

	echo "git2_${urlSite}${urlPath//\//.}"
}
srcFilename=$(git_base_name $srcUrl)

git_tarball_name() {
	local gitUrl="${1:-$srcUrl}"
	local srcFilename=$(git_base_name $gitUrl)
	
	# LZO would be better, but needs LZOP to be preinstalled, as we can't define fetch dependencies...
	echo $srcFilename.tar.gz
	# echo $srcFilename.tar.lzo 
}

git_create_tarball() {
	local gitUrl="${1:-$srcUrl}"
	local targetDir="${2:-$downloadDir}"
	local srcFilename="${3:-$(git_base_name $gitUrl)}"
	local repoDir="$targetDir/git-repos/$srcFilename"
	local tarball="$targetDir/$(git_tarball_name $gitUrl $srcFilename)"

	if [ -r "$tarball" ]; then
		# Here we check if the remote origins (refs/remotes) are more
		# recent than our tarball and only rebuild the tarball if needed.
		# If we use this git repo as a development repo, we must also
		# regenerate the tarball after a commit or any other operation
		# changing a ref in the repo, that is creating a log entry
		# When updating a tarball, don't delete it first as it may hold a lock...
		for ref in $(find $repoDir/refs $repoDir/logs/refs -type f); do
			if [ "$ref" -nt "$tarball" ]; then
				build_tarball "$tarball" "$repoDir" .
				break
			fi
		done
	else
		build_tarball "$tarball" "$repoDir" .
	fi
}

git_set_outstamp_from_srcRev() {
	local srcRev="$1"
	local statefile="$2"
	
	if [ -n "$srcRev" -a -z "${srcRev//[A-Fa-f0-9]}" ]; then
		# Revision is a commit, will never change
		echo "$srcRev" > "$stateFile.outstamp"
	else
		# Not a commit: must redo import each time
		poison_dependent_tasks
	fi
}

git_check() {
	local gitUrl="${1:-$srcUrl}"
	local targetDir="${2:-$downloadDir}"
	local srcFilename="${3:-$(git_base_name $gitUrl)}"
	local srcFile="$targetDir/$srcFilename"	
	local stateFile="$(recipe_task_state_file)"
	local repoDir="$targetDir/git-repos/$srcFilename"

	if [ ! -r "$srcFile" ]; then return 1; fi
	oldRepo="$(cat "$srcFile")"
	if [ "$repoDir" != "$oldRepo" ]; then return 1; fi

	if [ ! -r "$srcFile.done" ]; then return 1; fi
	local oldUrl="$(cat "$srcFile.done")"
	if [ "$gitUrl" != "$oldUrl" ]; then return 1; fi

	if [ ! -d "$repoDir" ]; then
		if [ ! -r "$srcFile.tar.gz" ]; then return 1; fi
		mkdir -p "$repoDir"
		log_exec_locked "$srcFile.tar.gz" tar xf "$srcFile.tar.gz" -C "$repoDir"
	fi

	git_set_outstamp_from_srcRev "$srcRev" "$stateFile"
	return 0 # Update and checksum will be done in the import phase
}

git_clone_url() {
	local gitUrl="$1"
	local repoDir="$2"

	if log_run git clone --bare --progress "$gitUrl" "$repoDir"; then
		return 0
	fi
	
	# Some git repositories now reject git:// URIs, so try also the equivalent https:// URI
	# Note this may be useless, dating from a time we do not obey srcProtocol
	local altUrl="http://${gitUrl#git://}"
	if log_run git clone --bare --progress "$altUrl" "$repoDir"; then
		return 0
	fi
	
	# Also try with explicit https
	altUrl="https://${gitUrl#git://}"
	if log_run git clone --bare --progress "$altUrl" "$repoDir"; then
		return 0
	fi
	
	# Definitely does not work...
	return 1
}

git_fetch() {
	local gitUrl="${1:-$srcUrl}"
	local targetDir="${2:-$downloadDir}"
	local srcFilename="${3:-$(git_base_name $gitUrl)}"
	local tarballName="$(git_tarball_name $gitUrl $srcFilename)"
	local srcFile="$targetDir/$srcFilename"	
	local stateFile="$(recipe_task_state_file)"
	local repoDir="$targetDir/git-repos/$srcFilename"

	# First check if we already fetch it and, in this case, just update the tarball if needed
	if git_check "$gitUrl" "$targetDir" "$srcFilename"; then
		if log_run git -C "$repoDir" --bare remote update; then
			git_create_tarball "$gitUrl" "$targetDir" "$srcFilename"
		fi
		return
	fi
	
	rm -rf "$repoDir" "$srcFile"
	mkdir -p "$targetDir/git-repos"
	
	# First try to fetch from a mirror tarball, as git cloning may be
	# quite slow when the repository is huge
	local tarballUrl="${gitUrl%/*}/$tarballName"
	local fetched=$(wget_from_mirror "$downloadDir/$tarballName" "$tarballUrl" true true $quiet)
	if [ -n "$fetched" -a -r "$fetched" ]; then
		#
		# Note here we fetch a file from the internet without checking its hash;
		# However, with git, this is not a problem, as long as we specify an explicit
		# commit as, in this case, the commit is the hash of the whole fetched source.
		#
		mkdir -p "$repoDir"
		log_exec_locked $fetched tar xf $fetched -C "$repoDir"
		
		# However we can't trust the mirror to be up to date, so update our copy
		log_run git -C "$repoDir" --bare remote update
	else
		# first look for a GIT mirror, as some official GIT servers are sometimes overloaded...
		# We may match mirrors to the URI server and path
		local mirror_spec mirrorNo mirrorStub mirrorPath mirrorBase mirrorUrl fetched=false
		
		# First sort the mirror list to have the longest stubs first
		mirrorNo=0
		for mirror_spec in ${DEFAULT_GIT_MIRRORS} ${EXTRA_GIT_MIRRORS}; do
			let mirrorNo+=1
			mirrorStub="${mirror_spec%%|*}"
			mirrorBase="${mirror_spec#*|}"
			echo ${#mirrorStub} "$mirrorNo" "$mirrorStub" "$mirrorBase"
		done | sort -k1rn,1 -k2n,2 > $tmpdir/git-mirrors

		while read mirror_spec mirrorNo mirrorStub mirrorBase; do
			mirrorPath=${gitUrl#${mirrorStub%/}/}
			if [ "${gitUrl}" != "${mirrorPath}" ]; then
				mirrorUrl="${mirrorBase%/}/${mirrorPath}"
				if git_clone_url "$mirrorUrl" "$repoDir"; then
					fetched=true
					break
				fi
			fi
		done < $tmpdir/git-mirrors
		
		if ! $fetched && ! git_clone_url "$gitUrl" "$repoDir"; then
			# We already tried mirrors, so no other possibility...
			remove_task_output_state
			return 1
		fi
	fi
	echo "$gitUrl" > "$srcFile.done"
	echo "$repoDir" > "$srcFile"
	git_set_outstamp_from_srcRev "$srcRev" "$stateFile"
	git_create_tarball "$gitUrl" "$targetDir" "$srcFilename"
}

git_import() {
	local gitUrl="${1:-$srcUrl}"
	local gitFile="${2:-$srcFile}"
	local destDir="${3:-$srcDir}"
	local here=$(pwd)
	local stateFile=$(recipe_task_state_file)
	local fetchStateFile=$(recipe_task_state_file fetch)
	local parentStateFileDir=$(dirname "$stateFile")

	if [ ! -r "$srcFile" ]; then
		rm -rf "$srcFile.done" "$fetchStateFile"
		fatal 1 "Fetch seems to have failed..."
	fi
	
	local repoDir=$(cat "$gitFile")
	note "repoDir='$repoDir'"
	if [ ! -d "$repoDir" ]; then
		rm -rf "$gitFile" "$gitFile.done" "$fetchStateFile.stamp"
		fatal 2 "Fetch seems to have been corrupted"
	fi
	log repoDir=$repoDir
	
	local optRev optDate
	if [ -n "$srcRev" ]; then
		optRev="$srcRev"
	else 
		if [ -n "$srcBranch" ]; then
			optRev="$srcBranch"
		else
			optRev="master"
		fi
	fi
	if [ -n "$srcDate" -a -n "${optRev//[A-Fa-f0-9]}" ]; then
		optDate=" '--before=$srcDate'"
	fi
	log optRev=$optRev

	mkdir -p $parentStateFileDir
	
	local myRev="${optRev}${optDate}"
	local oldRev="$myRev"
	if [ -r "$stateFile.revno" ]; then
		oldRev=$(cat "$stateFile.revno")
		if [ "$myRev" != "$oldRev" ]; then
			echo "$myRev" > "$stateFile.revno"
		fi
	else
		echo "$myRev" > "$stateFile.revno"
	fi
	
	if [ -n "${optRev//[A-Fa-f0-9]}" ]; then # Not a commit: update from remote
		note "Not fetching a commit, mustupdate our local bare clone"
		log_exec git -C "$repoDir" --bare remote update
		# Get the proper commit
		optRev=$(git -C "$repoDir" --bare rev-list -n 1 $optDate $optRev)
		echo "$optRev" > "$stateFile.revno"
	fi
	
	if [ ! -r "$destDir/.git" ]; then
		# Do not remove destDir, as this may also remove parentStateFileDir
		# which contains valuable data
		# The problem is that, then, we can no more do a plain git clone here
		# as cloning needs a non-existing directory. So we clone in a git-clone
		# subdirectory and move everything back to $destDir manually...
		rm -rf "$destDir/git-clone"
		if log_run git clone "$repoDir" "$destDir/git-clone"; then
			cd "$destDir"
			log "Moving content of git-clone in $destDir"
			local gitfile dstfile
			for gitfile in git-clone/* git-clone/.[^.]* git-clone/..?*; do
				if [  ! -r $gitfile ]; then continue; fi
				dstfile=$(basename $gitfile)
				rm -rf $dstfile
				mv $gitfile $dstfile
			done
			# git-clone should now be empty, just remove it
			if is_tree_not_empty git-clone; then
				fatal "Cloned directory '$destDir/git-clone' contains unmovable files..."
			fi
			log_exec rm -rf git-clone
		else
			rm -f "$stateFile.outstamp"
			fatal 3 "Can't initialize working directory '$destDir' from '$gitUrl'"
		fi
	fi
	
	cd "$destDir"
	if [ -n "${optRev//[A-Fa-f0-9]}" -o "$optRev" != "$oldRev" ]; then 
		# Not a commit, or changed: update the cloned workspace;
		# note the local bare clone was already update above
		#log_exec git -C "$repodir" remote update
		log_exec git remote update
	fi

	if log_run git checkout $optRev; then
		git log -n 1 '--pretty=tformat:%H' > "$stateFile.outstamp"
	else
		rm -f "$stateFile.outstamp"
		fatal 4 "Can't check out $optRev$optDate from '$gitUrl'"
	fi
}
