######################################################################################################################
#                                                                                                                    #
#                                                   SVN utilities                                                    #
#                                                                                                                    #
######################################################################################################################

generic_svn_opts=" --non-interactive --trust-server-cert"

svn_get_revSpec() {
	local svnRev="${1:-$srcRev}"
	local svnDate="${2:-$srcDate}"
	local revSpec 
	if [ -n "$svnRev" ]; then
		revSpec="$(trim $svnRev)"
		if [ -n "${srcRev//[0-9]}" ]; then
			error "Incorrect Subversion revision number '$svnRev'"
		fi
	elif [ -n "$svnDate" ]; then # Date format is quite free"
		revSpec=$(date --date="$svnDate" -u +@%s)
		if [ $? -ne 0 ]; then
			error "Incorrect date specification '$svnDate'"
		fi
	fi
	if [ -z "$revSpec" ]; then
		revSpec="HEAD"
	fi
	echo "$revSpec"
}

svn_get_revOpt() {
	local revSpec="${1:-$(svn_get_revSpec)}"
	local revOpt 
	
	case $revSpec in
		HEAD)
			revOpt=
			;;
		@*)
			revOpt="-r '{$(date --date="$revSpec" --rfc-3339=seconds)}'"
			;;
		*)
			revOpt="-r $revSpec"
			;;
	esac
	echo "$revOpt"
}

# The name is compatible with the Yocto name to benefit from their mirrors...
svn_base_name() {
	local svnUrl="${1:-$srcUrl}"
	local revSpec="${2:-$(svn_get_revSpec)}"
	local urlSite="${svnUrl#*://}"
	local urlPath="/${urlSite#*/}"
	urlSite="${urlSite%%/*}"

	if [ -n "$srcModule" ]; then
		echo "${srcModule//\//.}_${urlSite}_${urlPath//\//.}_${revSpec}_"
	else
		echo "${urlSite}_${urlPath//\//.}_${revSpec}_"
	fi	
}
srcFilename=$(svn_base_name $srcUrl)

svn_tarball_name() {
	local svnUrl="${1:-$srcUrl}"
	local srcFilename="${2:-$(svn_base_name $svnUrl)}"
	
	echo $srcFilename.tar.gz
}

svn_create_tarball() {
	local svnUrl="${1:-$srcUrl}"
	local targetDir="${2:-$downloadDir}"
	local srcFilename="${3:-$(svn_base_name $svnUrl)}"
	local dirName="${4:-$srcModule}"
	local repoDir="$targetDir/svn-repos/$srcFilename"
	local tarball="$targetDir/$(svn_tarball_name $svnUrl $srcFilename)"

	dirName=${dirName%/trunk}
	dirName=$(basename $dirName)
	if [ -z "$dirName" ]; then dirName="${PACKAGE_NAME^}"; fi
	build_tarball $(svn_tarball_name $svnUrl $srcFilename) \
	              "$targetDir/svn-repos" \
	              "--transform=s:$srcFilename:$dirName:" \
	              "$srcFilename"
}

svn_check() {
	local svnUrl="${1:-$srcUrl}"
	local targetDir="${2:-$downloadDir}"
	local srcFilename="${3:-$(svn_base_name $svnUrl)}"
	local downloadDir="$2"
	local srcFilename="$3"
	local srcFile="$downloadDir/$srcFilename"	
	local fetchDoneFile="$srcFile.done"
	local repoDir="$downloadDir/svn-repos/$srcFilename"
	local revSpec="$(svn_get_revSpec)"
	
	if [ "$revSpec" = "HEAD" ]; then return 1; fi

	if [ ! -r "$srcFile" ]; then return 1; fi
	oldRepo="$(cat "$srcFile")"
	if [ "$repoDir" != "$oldRepo" ]; then return 1; fi
	
	if [ ! -d "$repoDir" ]; then return 1; fi
	
	if [ ! -r "$fetchDoneFile" ]; then return 1; fi
	local oldUrl="$(cat "$fetchDoneFile")"
	if [ "$srcUrl+$revSpec" != "$oldUrl" ]; then return 1; fi

	return 0 
}

svn_clone() {
	local srcUrl="$1" 
	local repoDir="$2"
	local revSpec="$3"
	local revOpt=$(svn_get_revOpt $revSpec)
	
	if [ -n "$srcModule" ]; then
		srcUrl="$srcUrl/$srcModule"
	fi

	log_run rm -rf $repoDir # Clean as, otherwise, we may not be able to checkout
	log_run svn $generic_svn_opts -q co ${revOpt} $srcUrl $repoDir
	if [ $? -ne 0 ]; then return 1; fi

	return 0
}

svn_fetch() {
	local srcUrl="$1" 
	local downloadDir="$2"
	local srcFilename="$(svn_base_name $srcUrl)"
	local srcFile="$downloadDir/$srcFilename"	
	local fetchDoneFile="$srcFile.done"
	local repoDir="$downloadDir/svn-repos/$srcFilename"
	local revSpec=$(svn_get_revSpec)
	local fetchStateFile="$(recipe_task_state_file)"

	if svn_check "$srcUrl" "$downloadDir" "$srcFilename"; then
		if [ -z "${revOpt}" ]; then
			# We have fetched the HEAD revision, so we may have to update it
			log_exec svn up ${revOpt} "${repoDir}"
		else
			# We have fetched a static revision, so no need to update it
			return
		fi
	else
		rm -rf "$repoDir" "$fetchDoneFile" "$srcFile"
		mkdir -p "$downloadDir/svn-repos"
		if svn_clone "$srcUrl" "$repoDir" "$revSpec"; then
			echo "$srcUrl+$revSpec" > "$fetchDoneFile"
			echo "$repoDir" > "$srcFile"
		elif [ "$revSpec" != "HEAD" ]; then
			# Can't clone the original svn repository, look for a mirror tarball
			# and, if we find one, just use it without looking at its log (we don't
			# have any, as we don't have access to the original repo
			local tarballName="$(svn_tarball_name $svnUrl)"
			local tarballUrl="${srcUrl%/*}/$tarballName"
			local fetched=$(wget_from_mirror "$downloadDir/$tarballName" "$tarballUrl" true true $quiet)
			if [ -n "$fetched" ]; then
				#
				# Note this is the only case where we fetch a file from the internet
				# without checking any kind of checksum hash... Perhaps we could
				# do such a check by carefully building the content in lexical order
				# and saving the MD5 and SHA256 checksums, but this does not allow to
				# fetch from non-SW4L managed mirrors...
				#
				mkdir -p "$repoDir"
				log_exec tar xf $fetched -C "$repoDir" --strip=1 
				echo "$srcUrl+$revSpec" > "$fetchDoneFile"
				echo "$repoDir" > "$srcFile"
				echo "$srcFilename" > "$fetchStateFile.stamp"
				return
			fi
		else
			error "Can't clone '$srcUrl' using subversion and can't look HEAD in mirrors"
		fi
	fi
	
	svn $generic_svn_opts log -l 1 $repoDir | sed -e '1d' -e '3,$d' > "$fetchStateFile.stamp.new"
	if cmp -s "$fetchStateFile.stamp" "$fetchStateFile.stamp.new"; then
		# No change in release, so no need to redo following steps
		rm -f "$fetchStateFile.stamp.new"
	else
		# Update the stamp file to invalidate next steps
		rm -f "$fetchStateFile.stamp"
		mv "$fetchStateFile.stamp.new" "$fetchStateFile.stamp"
		# and rebuild the tarball (this will be used if we want to create a mirror)
		svn_create_tarball $srcUrl $downloadDir $srcFilename $srcModule 
	fi
}

svn_import() {
	local srcUrl="$1" # unused here
	local srcFile="$2"	
	local destDir="$3" 
	local repoDir=$(cat "$srcFile")
	local srcFilename="$(basename $srcFile)"
	local downloadDir=$(dirname "$srcFile")
	local here="$(pwd)" 
	local fetchDoneFile="$srcFile.done"
	local fetchStateFile="$(recipe_task_state_file fetch)"
	
	if ! svn_check "$srcUrl" "$downloadDir" "$srcFilename"; then
		rm -rf "$srcFile" "$fetchDoneFile" "$fetchStateFile.stamp"
		fatal 1 "Fetch seems to have failed..."  
	fi

	log_exec mkdir -p "${destDir}"
	log "tar cf - -C $repoDir --exclude=.svn . | tar xf - -C $destDir"
	tar cf - -C $repoDir --exclude=.svn . | tar xf - -C $destDir
	local status=$?
	if [ $status -ne 0 ]; then
		error "Error copying '$repodir' in '$destDir'"
	fi
}
