######################################################################################################################
#                                                                                                                    #
#                                                     SD Card utilities                                              #
#                                                                                                                    #
######################################################################################################################

# Columns index for FLASHLAYOUT
FLASHLAYOUT_COL_TYPE=0
FLASHLAYOUT_COL_SIZE=1
FLASHLAYOUT_COL_FIRST_SECTOR=2
FLASHLAYOUT_COL_NB_PARTITIONS=3
FLASHLAYOUT_COL_IMG_FILENAME=4

# Columns index for PARTITION_data
PARTITION_COL_NAME=0
PARTITION_COL_START_SECTOR=1
PARTITION_COL_SIZE=2
PARTITION_COL_TYPE=3
PARTITION_COL_FILE=4
PARTITION_COL_FORMAT=5
PARTITION_COL_BOOTABLE=6
PARTITION_COL_IMAGE=7

# Section
SECTION_SIZE=512

# Scale factor to get valid image size
# Note: The image size has to be bigger than the folder size, it ensure no error,
#       The user can manually reduce image size afterward
: ${ROOTFS_SCALE_FACTOR:=1.3}

declare PARTITION_TABLE_TYPE
declare SDCARD_FILE_SIZE
declare FIRST_SECTOR
declare NB_PARTITIONS
declare SDCARD_FILENAME

declare -g -A PARTITION_data

here=$(pwd)
outputDir=$here/output

# Create temporary workdir
WORKDIR=$(mktemp -d -t tmp-XXXXXXXXXX --tmpdir=${outputDir})
trap "rm -rf $WORKDIR" EXIT

###########################################################################
# Get the tsv flashlayout file
###########################################################################
find_flashlayout_file() {
    # If TSVFILE variable is set, return it as flashlayout file    
    if [ ! -z ${TSVFILE+x} ]; then
    	if [ -f "${srcDir}/${TSVFILE}" ]; then
        	echo $TSVFILE
        else
        	warning "TSVFILE variable is set to \"${TSVFILE}\"; this file does not exist"
        fi
    fi    	
	
	message "Auto-detecting tsv layout file in rootfs project..."
    count=`ls -1 *.tsv 2>/dev/null | wc -l`
    if [ $count == 0 ]; then
        fatal 1 "No flashlayout TSV file found in '$srcDir'"
    fi

    # Search the first tsv file in project source directory    
    for filename in ${srcDir}/*.tsv; do
        echo $filename
    done
}

###########################################################################
# Read the tsv flash layout file and store info into PARTITION_data array
#    tsvFile        The TSV file to parse
###########################################################################
parse_flash_layout() {
    local tsvFile="$1"
    local lineNb=0
    local partNb=0
    declare -a flashlayout_line
    while IFS=$'\t' read -ra flashlayout_line; do

        # Skip comment lines and empty line
        [[ -z "${flashlayout_line[0]}" ]] && continue ;
        [[ "${flashlayout_line[0]}" =~ ^#.* ]] && continue ;
        
        case $lineNb in
        0)
            echo "Loading sdcard configuration ...";;
        1)
            PARTITION_TABLE_TYPE=${flashlayout_line[$FLASHLAYOUT_COL_TYPE]}
            SDCARD_FILE_SIZE=${flashlayout_line[$FLASHLAYOUT_COL_SIZE]}
            FIRST_SECTOR=${flashlayout_line[$FLASHLAYOUT_COL_FIRST_SECTOR]}
            NB_PARTITIONS=${flashlayout_line[$FLASHLAYOUT_COL_NB_PARTITIONS]}
            SDCARD_FILENAME=${flashlayout_line[$FLASHLAYOUT_COL_IMG_FILENAME]}
            log "Partition Table Configuration Found:"
            log "  Type=$PARTITION_TABLE_TYPE"
            log "  Size=$SDCARD_FILE_SIZE"
            log "  First-lba=$FIRST_SECTOR"
            log "  Number of partititons=$NB_PARTITIONS"
            log "====================================================================="
            ;;
        2)
            # Separator
            ;;
        3)
            # Partition header
            ;;
        *)        
            PARTITION_data[$partNb,$PARTITION_COL_NAME]=${flashlayout_line[$PARTITION_COL_NAME]}
            PARTITION_data[$partNb,$PARTITION_COL_START_SECTOR]=${flashlayout_line[$PARTITION_COL_START_SECTOR]}
            PARTITION_data[$partNb,$PARTITION_COL_SIZE]=${flashlayout_line[$PARTITION_COL_SIZE]}
            PARTITION_data[$partNb,$PARTITION_COL_TYPE]=${flashlayout_line[$PARTITION_COL_TYPE]}
            PARTITION_data[$partNb,$PARTITION_COL_FILE]=${flashlayout_line[$PARTITION_COL_FILE]}
            PARTITION_data[$partNb,$PARTITION_COL_FORMAT]=${flashlayout_line[$PARTITION_COL_FORMAT]}
            PARTITION_data[$partNb,$PARTITION_COL_BOOTABLE]=${flashlayout_line[$PARTITION_COL_BOOTABLE]}
            log "Partition $partNb:"
            log "  Name=${PARTITION_data[$partNb,$PARTITION_COL_NAME]}"
            log "  Start-sector=${PARTITION_data[$partNb,$PARTITION_COL_START_SECTOR]}"
            log "  Size=${PARTITION_data[$partNb,$PARTITION_COL_SIZE]}"
            log "  Type=${PARTITION_data[$partNb,$PARTITION_COL_TYPE]}"
            log "  File=${PARTITION_data[$partNb,$PARTITION_COL_FILE]}"
            log "  Format=${PARTITION_data[$partNb,$PARTITION_COL_FORMAT]}"
            log "  Bootable=${PARTITION_data[$partNb,$PARTITION_COL_BOOTABLE]}"
            log "====================================================================="
            partNb=$((partNb + 1))
            ;;
        esac
        # Increment line
        lineNb=$((lineNb + 1))
    done < "$tsvFile"

    # Set number of partition 
    NB_PARTITIONS=$partNb
}

###########################################################################
# Get the file size in sections
#    file        The path to the file
###########################################################################
get_file_size_in_sections() {
    local file=$1
    fileSize=$(stat -c%s $file)            
    fileSizeInSection=$((fileSize / SECTION_SIZE))
    echo $fileSizeInSection
}

###########################################################################
# Recalculate and adjust every partitions
###########################################################################
recalc_partitions() {
    for ((i=0;i<NB_PARTITIONS;i++))
    do
        # Partition start section
        case ${PARTITION_data[$i,$PARTITION_COL_START_SECTOR]} in
            auto)
                # If first sector is address is auto, put it in address 0x0
                if [ $i -eq 0 ]; then
                    PARTITION_data[$i,$PARTITION_COL_START_SECTOR]=${FIRST_SECTOR}
                else
                    # Calculate start section to put partition after previous partition
                    h=$((i - 1))
                    newStartSection=$((PARTITION_data[$h,$PARTITION_COL_START_SECTOR] + PARTITION_data[$h,$PARTITION_COL_SIZE] + 1))
                    PARTITION_data[$i,$PARTITION_COL_START_SECTOR]=${newStartSection}
                fi
                log "Partition '${PARTITION_data[$i,$PARTITION_COL_NAME]}' start sector is ${PARTITION_data[$i,$PARTITION_COL_START_SECTOR]}"
                ;;
            *) 
                # Keep value
                ;;
        esac

        # Partition size
        case ${PARTITION_data[$i,$PARTITION_COL_SIZE]} in
            auto)
                j=$((i + 1))
                # Calculate minimal size for the partition from partition file
                file="${PARTITION_data[$i,$PARTITION_COL_IMAGE]}"    
                newSize=$(get_file_size_in_sections $file)
                newEndSector=$((PARTITION_data[$i,$PARTITION_COL_START_SECTOR] + newSize))

                # No restriction if this partition is the last one or if the next partition start sector is "auto"
                if [ $j -le $NB_PARTITIONS ] && [ ${PARTITION_data[$j,$PARTITION_COL_START_SECTOR]} != "auto" ]; then
                    # Check is partition does not overlap next partition
                    if [ $newEndSector -gt ${PARTITION_data[$j,$PARTITION_COL_START_SECTOR]} ]; then
                        fatal 1 "Partition '${PARTITION_data[$j,$PARTITION_COL_NAME]}' size overlap partition '${PARTITION_data[$j,$PARTITION_COL_NAME]}', \
change '${PARTITION_data[$j,$PARTITION_COL_NAME]}' start sector or lower '${PARTITION_data[$j,$PARTITION_COL_NAME]}' size"
                    fi
                fi
                PARTITION_data[$i,$PARTITION_COL_SIZE]=${newSize}
                log "Partition '${PARTITION_data[$i,$PARTITION_COL_NAME]}' size is ${PARTITION_data[$i,$PARTITION_COL_SIZE]} sectors"
                ;;
            fill)
                # Calculate maximum size for the partition by grabbing extra empty space
                j=$((i + 1))
                if [ $j -lt $NB_PARTITIONS ]; then
                    if [ ${PARTITION_data[$j,$PARTITION_COL_START_SECTOR]} != "auto" ]; then
                        newSize=$((PARTITION_data[$j,$PARTITION_COL_START_SECTOR] - PARTITION_data[$i,$PARTITION_COL_START_SECTOR]))
                    else 
                        file="${PARTITION_data[$i,$PARTITION_COL_IMAGE]}"    
                        newSize=$(get_file_size_in_sections $file)
                    fi
                else
                    newSize=$(((SDCARD_FILE_SIZE*1000000/SECTION_SIZE) - PARTITION_data[$i,$PARTITION_COL_START_SECTOR]))
                fi
                PARTITION_data[$i,$PARTITION_COL_SIZE]=${newSize}
                
                log "Partition '${PARTITION_data[$i,$PARTITION_COL_NAME]}' size is ${PARTITION_data[$i,$PARTITION_COL_SIZE]} sectors"
                ;;
            *)
                # Keep value
                ;;
        esac
    done

}

##################################################
# Create sfdisk dump file for partition table and
# generate partition table
#     imgFile                sfdisk dump file to define disk image
#    tableType            table type (dos/gpt)
#    tableFirstSector    first sector offset (in sectors number)
#    nbPartitions        number of partitions and shadow partitions
##################################################
create_partition_table() {
    local imgFile="${1}"
    local tableType="${2,,}"
    local tableFirstSector="${3}"
    local nbPartitions="${4}"
    
    # Temporary dump file to describe table/partition info for sfdisk
    local dumpFile="${WORKDIR}/sfdisk.dump"
    
    if [ -z "$tableType" ]; then
        fatal 1 "Missing partition table type"
    fi
    
    # Create sfdisk dump file
    echo -e "label:${tableType}\nfirst-lba:${tableFirstSector}\n" > ${dumpFile}
    
    part_script=$(search_script partitioning/${tableType}_utils.sh)
    if [ -z "$part_script" ]; then
        fatal 1 "Cannot find handler to generate image with ${tableType} partition table"
    fi
    load_script_once "$part_script"
    
    # Create partition table
    for ((i=0;i<$NB_PARTITIONS;i++))
    do
        # Check if partition should be bootable
        local extraOpts=""
        if [ "${PARTITION_data[$i,$PARTITION_COL_BOOTABLE]}" == "Yes" ]; then
            extraOpts=", attrs=LegacyBIOSBootable, bootable"
        fi
        
        # Generated partition info in dump file
        ${tableType}_sfdisk_add_partition ${dumpFile} \
                    ${PARTITION_data[$i,$PARTITION_COL_NAME]} \
                    ${PARTITION_data[$i,$PARTITION_COL_TYPE],,} \
                    ${PARTITION_data[$i,$PARTITION_COL_START_SECTOR]} \
                    ${PARTITION_data[$i,$PARTITION_COL_SIZE]} \
                    "${extraOpts}"                                    
    done

    log_exec sfdisk ${imgFile} < ${dumpFile}
    
}

##################################################
# Create the partition image file
##################################################
create_partition_images() {
    imagesDir=${WORKDIR}
    for ((i=0;i<NB_PARTITIONS;i++))
    do
        partSrc="${PARTITION_data[$i,$PARTITION_COL_FILE]}"
        partType="${PARTITION_data[$i,$PARTITION_COL_FORMAT],,}"
        partName="${PARTITION_data[$i,$PARTITION_COL_NAME]}"
        partSizeSections="${PARTITION_data[$i,$PARTITION_COL_SIZE]}"
        partSize=$((partSizeSections / 2))             # nb_sector*sector_size/1024
        
        # If source is a file
        if [ -f "${partSrc}" ]; then
            # Copy source image into imagesDir
            log_exec cp ${partSrc} ${imagesDir}
            PARTITION_data[$i,$PARTITION_COL_IMAGE]=${imagesDir}/$(basename $partSrc)
        # Else If source is a directory
        elif [ -d "${partSrc}" ]; then
            # Generate image into imagesDir
            echo "generate_rootfs ${partSrc} ${partName} ${partSize}"
            generate_rootfs "${partSrc}" \
                            "${partType}" \
                            "${partName}" \
                            "${partName}"             
            PARTITION_data[$i,$PARTITION_COL_IMAGE]=${imagesDir}/${partName}.${partType}
        else
            fatal 3 "Partition ${partName} content source \"${partSrc}\" not found"
        fi
    done
}

##################################################
# Burn the partition image into the sdcard image file
#    sdimgFile        The sdcard image file path
##################################################
burn_partitions() {
    local sdimgFile=$1
    
    # For each partition, check if partition image size fit in partition size 
    # defined in the table then copy the content into the sdcard image at chosen address
    for ((i=0;i<$NB_PARTITIONS;i++)); do
        partSize="${PARTITION_data[$i,$PARTITION_COL_SIZE]}"
        file=$(eval echo "${PARTITION_data[$i,$PARTITION_COL_IMAGE]}")
        imgSize=$(get_file_size_in_sections $file)
        
        # Check if image fits its partition
        if [ "${imgSize}" -gt "${partSize}" ]; then
            fatal 2 "$file cannot fit into ${PARTITION_data[$i,$PARTITION_COL_NAME]} (${imgSize} sectors vs ${partSize} sectors)"
        fi
        log_exec dd if="${PARTITION_data[$i,$PARTITION_COL_IMAGE]}" of="${sdimgFile}" \
                    conv=fdatasync,notrunc seek="${PARTITION_data[$i,$PARTITION_COL_START_SECTOR]}" bs=$SECTION_SIZE
    done
}

##################################################
# Function to call to create SD Card image from 
# layout TSV file and populated it with generated rootfs output
##################################################
do_create_sdimg() {
    # Find the tsv flashlayout file

    local tsvFile=$(find_flashlayout_file)
    log "Flashlayout TSV file detected: ${tsvFile}"
    
    # Load SD Card image information for TSV file
    log "Parsing SD Card layout file"
    parse_flash_layout $tsvFile
    
    # Remove old image if exists
    : {SDCARD_FILENAME:=img.sdcard}
    sdimgFile=${outputDir}/${SDCARD_FILENAME}
    rm -f ${SDCARD_FILENAME}
    
    # Create partition images
    log "Generating partitions..."
    create_partition_images
    
    # Initialize sdcard image file
    log "Initialize sdcard image file: ${sdimgFile} of ${SDCARD_FILE_SIZE}MB..."
    log_exec dd if=/dev/zero of=${sdimgFile} bs=1024 count=0 seek=${SDCARD_FILE_SIZE}K
    
    # Recalculate partitions sizes if needed
    recalc_partitions

    # Create partition table
    log "Create partitions table..."
    create_partition_table ${sdimgFile} $PARTITION_TABLE_TYPE $FIRST_SECTOR $nbPartitions
    
    # Burn partitions
    log "Burn partitions into sdcard image..."
    burn_partitions ${sdimgFile}

    # Generate outstamp on tsvfile to force task rerun if tsv file was modified
    generate_task_output_state $(sha1sum ${tsvFile})
}
