#
# Split installation into tarball feeds
# - base     : items that must always be installed
# - dev      : items for development: header files, dynamic libraries links
# - staticdev: static libraries
# - dbg      : items for delivery with debug feature unstripped
# - doc      : items for delivery with generated documentations
# - locale   : items related to internationalization
#
# WARNING: This is sourced inside a function, so all variables
# created by the declare builtin are implicitly local to this
# (short-lived) function, so you MUST declare them with the -g
# flag; normal variables being implicitly global have no problem.

: ${debug_package_split:=false}
# debug_package_split=true
$debug_package_split && print_traces split

split_dbg() {
	dbg split "$@"
}
split_dbg_cat() {
	dbg_cat split "$@"
}
split_dbg "Starting split step..."

# Give default values to parameters (:= if null or unset, = if only unset, variables must be defined)

: ${splitted_packages:=$package_name}
: ${sub_packages:="src doc config locale i18n shlibs staticdev dev dbg test base"}
: ${needed_subpackages:="staticdev dev dbg base"}
: ${unsplitted_shlibs:=libc libm}

set -o noglob
: ${pattern_base:="/bin /sbin /usr/bin /usr/sbin /usr/libexec /usr/share"}
: ${pattern_config:="/etc /usr/etc /usr/*/etc"}
: ${pattern_shlibs:="/lib/*.so /usr/lib/*.so /lib/*.so-* /usr/lib/*.so-* /lib/*.so.* /usr/lib/*.so.*"}
: ${pattern_locale:="/usr/share/locale"}
: ${pattern_i18n:="/usr/share/i18n"}
: ${pattern_staticdev:="/lib/*.a /usr/lib/*.a /usr/*/lib/*.a"}
: ${pattern_dev:="/lib/*.o /usr/lib/*.o /usr/*/lib/*.o /lib/*.la /usr/lib/*.la /usr/*/lib/*.la /usr/include /lib/pkgconfig /usr/lib/pkgconfig /lib/pkg-config /usr/lib/pkg-config /usr/share/pkgconfig /usr/share/aclocal"}
: ${pattern_dbg:="/bin/.debug /lib/.debug /usr/bin/.debug /usr/lib/.debug"}
: ${pattern_doc:="/usr/share/doc /usr/share/man /usr/share/info"}
set +o noglob

# Note that here we are only interested by target items, not native ones, and
# these will all go in the target feeds, leaving the native feed empty
if $recipe_is_native; then
	log "Nothing to split; native-only package"
	exit 0
fi

# Note tempDir, like any other xxxDir must be an absolute path 
# For the "cpio -pdlm" commands below don't have to copy anything, ensure
# we stay on the same device so that hard links work
tempDir=$workDir/_split.$$
record_cleanup "silent_do_locked $installDir rm -rf $tempDir"
rm -rf $tempDir

installTempDir=$tempDir/install
mkdir -p $installTempDir
copy_tree_excluding -d $installDir $installTempDir
cd $installTempDir

# Adjust the libtool scripts (although seldom used) and the DT_RPATH/DT_RUNPATH
adjust_libtool_scripts_and_rpath_for_packaging -l $installDir '.'

# Generate in a temporary directory, before moving in feedsDir
metadataTempDir="$tempDir/feeds/$metadataSubDir"
contentsTempDir="$tempDir/feeds/$contentsSubDir"
packagesTempDir="$tempDir/feeds/$packagesSubDir"
tarballsTempDir="$tempDir/feeds/$tarballsSubDir"

global_locales_file="${metadataDir%/*}/all.locales"

mkdir -p $contentsTempDir $packagesTempDir $tarballsTempDir $metadataTempDir
allContentsDir=$feedsDir/contents/$package_arch

base_bin_packages=
forced_bin_packages=
other_bin_packages=

# This script may be sourced in a function; without -g these variables would be local
declare -g -A subpkgs # Subpackages declared for a given base package
declare -g -A basepkg # Base package of a given subpackage
declare -g -A pkgkind # Kind (i.e. subpkg) of a given package
declare -g -A pkgname # Variable name of the package

empty_dir_marker="empty-dir-marker-to-be-deleted"

do_buildstep_prepend

# Generate list of packages and package content

check_package_base() {
	local pkg="$1"
	local isbase=true
	local base sub
	
	if [ "$pkg" != "$package_name" -a "${pkg%-*}" != "$pkg" ]; then
		for sub in $sub_packages; do
			if [ $sub = "base" ]; then continue; fi
			base=${pkg%-$sub}
			if [ "$base" = "$pkg" ]; then continue; fi
			split_dbg "Record '$pkg' is a sub-package of '$base'"
			if ! [[ " ${subpkgs[$base]} " =~ " $sub " ]]; then
				subpkgs[$base]="${subpkgs[$base]} $sub"
			fi
			basepkg[$pkg]="$base"
			pkgkind[$pkg]="$sub"
			isbase=false
			split_dbg "Found '$base' as base package for '$pkg' (1)"
			break
		done
		if $isbase; then
			# Check if there is another package with the same "base" root
			base=$pkg
			while true; do
				sub=${base%-*} # Suppress the last dash-delimited component
				if [ "$sub" = "$base" ]; then break; fi
				base=$sub
				if [[ " $splitted_packages $package_name " =~ " $base " ]]; then
					# This is a non-standard sub-package of a defined base package
					sub=${pkg#$base-}
					isbase=false
					split_dbg "Found '$base' as base package for '$pkg' (2)"
					break
				elif [[ " $splitted_packages $package_name " =~ " $base-" ]]; then
					# This maybe a non-standard sub-package of a possibly not yet knowned base package
					for sub in $splitted_packages $package_name; do
						if [ "$sub" = "$pkg" ]; then continue; fi
						if [[ " $sub " =~ " $base-" ]]; then
							# There are several sub-packages of base, so go ahead		
							sub=${pkg#$base-}
							isbase=false
							split_dbg "Found '$base' as base package for '$pkg' (3)"
							break
						fi
					done
				fi
				$isbase || break
			done
		fi
	fi
	if $isbase; then
		# This is really a plain base package
		subpkgs[$pkg]="${subpkgs[$pkg]} base"
		basepkg[$pkg]=
		pkgkind[$pkg]="base"
	else
		if [ -z "${subpkgs[$base]}" ]; then
			split_dbg "Seems that '$pkg' is a sub-package of '$base'; check it is a base..."
			check_package_base $base
		fi
		# Now record its new sub-package
		if ! [[ " ${subpkgs[$base]} " =~ " $sub " ]]; then
			subpkgs[$base]="${subpkgs[$base]} $sub"
		fi
		basepkg[$pkg]="$base"
		pkgkind[$pkg]="$sub"
	fi

	# Debug trace
	if $isbase; then
		split_dbg ">>>>>> $pkg"
	else
		split_dbg "       $pkg - $base"
	fi
	# End debug trace

	var_name=$(shell_variable_name $pkg)
	pkgname[$pkg]="$var_name"
}

#
# Generate package list, updating global array variables
# and creating splitted_packages if needed
#
generate_package_list() {
	local pkg
	
	if [ "$splitted_packages" = "$package_name" -o -z "$splitted_packages" ]; then
		local var_name=$(shell_variable_name "$package_name")
		splitted_packages=
		for pkg in $sub_packages; do
			if [ "$pkg" = "base" ]; then
				splitted_packages="${splitted_packages# } $package_name"
				subpkgs[$package_name]="${subpkgs[$package_name]# } base"
				basepkg[$package_name]=
				pkgkind[$package_name]="base"
				pkgname[$package_name]="$var_name"
			else
				splitted_packages="${splitted_packages# } $package_name-$pkg"
				subpkgs[$package_name]="${subpkgs[$package_name]# } $pkg"
				basepkg[$package_name-$pkg]="$package_name"
				pkgkind[$package_name-$pkg]="$pkg"
				pkg_name=${pkg//+/P}
				pkgname[$package_name-$pkg]="$(shell_variable_name $package_name-$pkg)"
			fi
		done
	else
		if [ "$NOAUTOPACKAGEDEBUG" != "1" ]; then
			if ! [[ " $splitted_packages " =~ " $package_name-dbg " ]]; then
				splitted_packages="${splitted_packages# } $package_name-dbg"
			fi
		fi
		
		if [ "$INHIBIT_PACKAGE_DEBUG_SPLIT" != "1" ]; then
			local var_name=$(shell_variable_name "pattern_$package_name-src")
			if [ "$PACKAGE_DEBUG_SPLIT_STYLE" = "debug-with-srcpkg" -o -z "$PACKAGE_DEBUG_SPLIT_STYLE" ]; then
				if ! [[ " $splitted_packages " =~ " $package_name-src " ]]; then
					splitted_packages="${splitted_packages# } $package_name-src"
				fi
				eval ${var_name}_src="\"\$${var_name}_src /usr/src/debug\""
			fi
			# Add all .debug directories to the main -dbg package 
			local dbgdir
			local pattern_var="${var_name}_dbg"
			for dbgdir in $(find . -type d -name .debug); do
				dbgdir="${dbgdir#./}"
				if [[ " $pattern_var " =~ " $dbgdir " ]]; then continue; fi
				if [[ " $pattern_var " =~ " /$dbgdir " ]]; then continue; fi
				eval $pattern_var="\"\$$pattern_var $dbgdir\""
			done
		fi
	
		for pkg in $splitted_packages; do
			check_package_base $pkg
		done
	fi
	
}

#
# Add the missing needed base or sub packages (will be
# generated as empty) creating the global variable all_packages
#
add_missing_packages() {
	all_packages=" $splitted_packages "
	for base in ${!subpkgs[@]}; do
		# Base is checked first so that if we need a subpackage, the base is there already
		for needed in "base" $needed_subpackages; do
			if [[ " ${subpkgs[$base]} " =~ " $needed " ]]; then
				continue
			fi
			if [ "$needed" = "base" ]; then 
				# Add it after the last sub-package
				split_dbg "Add missing base package '$base' due to subpackages '${subpkgs[$base]}'"
				lastsub=${subpkgs[$base]##* }
				all_packages=${all_packages/ $base-$lastsub / $base-$lastsub $base }
				subpkgs[$base]="${subpkgs[$base]} base"
				basepkg[$base]=
				pkgkind[$base]="base"
				var_name=$(shell_variable_name $base)
				pkgname[$base]="$var_name"
				split_dbg "Variable name for '$base' is '$var_name'"
			else
				# Add it before its base
				split_dbg "Add missing subpackage '$base-$needed'"
				all_packages=${all_packages/ $base / $base-$needed $base }
				subpkgs[$base]="${subpkgs[$base]} $needed"
				basepkg[$base-$needed]="$base"
				pkgkind[$base-$needed]="$needed"
				var_name=$(shell_variable_name $base-$needed)
				pkgname[$base-$needed]="$var_name"
				split_dbg "Variable name for '$base-$needed' is '$var_name'"
			fi
		done
	done
}

#
# Be careful that
#	1)	source packages must be first
#	2)	debug packages must be after source but before other packages
#	3)	all other packages must be kept in the order they were specified
#
declare -g -a sorted_packages # This script may be sourced in a function; without -g the variable would be local
sort_packages() {
	for pkg in $all_packages; do
		if [[ "$pkg " =~ "-src " ]]; then
			if [[ " $sorted_packages " =~ " $pkg " ]]; then
				warning "Package '$pkg' specified twice; as this would cause errors, ignore the second one"
				continue
			fi
			sorted_packages+=( "$pkg" )
		fi
	done
	for pkg in $all_packages; do
		if [[ "$pkg " =~ "-dbg " ]]; then
			sorted_packages+=( "$pkg" )
			if [[ " $sorted_packages " =~ " $pkg " ]]; then
				warning "Package '$pkg' specified twice; as this would cause errors, ignore the second one"
				continue
			fi
		fi
	done
	for pkg in $all_packages; do
		if [[ "$pkg " =~ "-src " ]] || [[ "$pkg " =~ "-dbg " ]]; then
			continue;
		fi
		if [[ " $sorted_packages " =~ " $pkg " ]]; then
			warning "Package '$pkg' specified twice; as this would cause errors, ignore the second one"
			continue
		fi
		sorted_packages+=( "$pkg" )
	done
}

duplicate_metadata() {
	local base_package="$1"
	local new_package="$2"
	local message="$3"

	for spec in SUMMARY DESCRIPTION SECTION HOMEPAGE LICENSE RSUGGESTS RCONFLICTS RREPLACES; do
		local lc_spec="${spec,,}"
		if [ -r "$metadataTempDir/$base_package.$lc_spec" -a ! -r "$metadataTempDir/$new_package.$lc_spec" ]; then
			if [ $spec = DESCRIPTION ]; then
				cp $metadataTempDir/$base_package.$lc_spec $metadataTempDir/$new_package.$lc_spec
				echo "This package contains the $message." >> $metadataTempDir/$new_package.$lc_spec
			elif [ $spec = SUMMARY ]; then
				echo "${message^}. " > $metadataTempDir/$new_package.$lc_spec
				cat $metadataTempDir/$base_package.$lc_spec >> $metadataTempDir/$new_package.$lc_spec
			else
				cp $metadataTempDir/$base_package.$lc_spec $metadataTempDir/$new_package.$lc_spec
			fi
		fi
	done
}

generate_preinst() {
	local bin_package="$1"
	local package="$2"
	local suffix=$(shell_variable_name $package)
	spec_groupadd_var=GROUPADD_PARAM_${suffix}
	spec_useradd_var=USERADD_PARAM_${suffix}
	spec_groupmems_var=GROUPMEMS_PARAM_${suffix}
	val_groupadd_set=${!spec_groupadd_var}
	val_useradd_set=${!spec_useradd_var}
	val_groupmems_set=${!spec_groupmems_var}
	if [ -n "$val_groupadd_set" -o -n "$val_useradd_set" -o -n "$val_groupmems_set" ]; then
		install -m 777 ${script_lib_dir}/template.preinst $metadataTempDir/${bin_package}.preinst
		sed -i -e "s,@GROUPADD_PARAM@,${val_groupadd_set},g" \
			-e "s,@USERADD_PARAM@,${val_useradd_set},g" \
			-e "s,@GROUPMEMS_PARAM@,${val_groupmems_set},g" $metadataTempDir/${bin_package}.preinst
	fi
}

build_package_tarball() {
	local bin_package="$1"
	local file_list="$2"
	local package="$3"
	local pkg="$4"
	local pkgTarball=${bin_package}.tar.bz2
	local dir_list="/tmp/dir-list.$$"
	local dir

	if [ -s "$file_list" ]; then
		log "Generating ${package} ${pkg^^} package feed: '$bin_package'"
		split_dbg_cat "Files_list:" $file_list
		for file in $(cat "$file_list"); do
			dir=$(dirname "$file")
			while [ "$dir" != "." ]; do
				echo "$dir"
				dir=$(dirname "$dir")
			done
		done | sort | uniq > "$dir_list"
		if grep -q "/$empty_dir_marker" "$file_list" 2> /dev/null; then
			grep -v "/$empty_dir_marker" "$file_list" > "$file_list.clean" 
			if [ -r "$file_list.clean" ]; then
				rm -f "$file_list"
				mv "$file_list.clean" "$file_list"
				echo rm -f "$file_list.clean"
			fi
			split_dbg_cat "Cleaned-up files list:" "$file_list"
		fi
		split_dbg_cat "Directories_list:" "$dir_list"
	else
		log "Force generation of empty ${package} ${pkg^^} package feed: '$bin_package'"
	fi

	if [ -s "$dir_list" ]; then
		log_exec_locked $installDir "tar -cjf '$tarballsTempDir/$pkgTarball' --no-recursion --files-from='$dir_list' --files-from='$file_list'"
		# Remove the temporary file to avoid unwanted re-use
		rm -f $dir_list	
	else
		log_exec_locked $installDir "tar -cjf '$tarballsTempDir/$pkgTarball' --files-from='$file_list'"
	fi

	echo "$package_id" > "$packagesTempDir/$bin_package"
	touch "$metadataTempDir/$recipe_id.manifest"
	echo "$bin_package" >> "$metadataTempDir/$recipe_id.manifest"

	# Remove the packaged files so that a file is only packaged once; need
	# to lock the installDir, due to the shallow copy (beware there may be
	# a LOT of files to delete...)
	silent_run_locked $installDir "(cat $file_list | xargs -n64 rm -f)"
}

append_dependencies() {
	local depfile="$1"
	local package dep
	package=${depfile##*/}
	package=${package%.*}
	touch "$depfile"
	touch "$depfile.tmp"
	split_dbg "Append dependencies to '$depfile' (package is '$package')"
	cat - "$depfile" | grep -v '^$' | sort | uniq | while read dep; do
		if [ "$dep" = "$package" ]; then 
			split_dbg "    Skip package: '$dep' == '$package'"
			continue
		fi
		echo "$dep" >> "$depfile.tmp"
	done
	if [ -s "$depfile.tmp" ]; then
		rm -f "$depfile"
		mv "$depfile.tmp" "$depfile"
	else
		rm -f "$depfile.tmp" "$depfile"
	fi
}

print_dependencies() {
	local bin_package="$1"
	local package_name=${bin_package%%_*}
	local package_vers=${bin_package#*_}
	shift
	
	local kind="$1"
	local ekind
	shift

	local subpkgs="$1"
	shift
	
	local rdeps="$*"

	local subpkg subpkg_dep pkg rdep rdep_name rdep_vers sep
	local pkgsub="${pkgkind[${package_name}]}"
	local subpkg_depvar="subpkg_${kind^^}_${pkgsub//-/_}"

	for subpkg_dep in ${!subpkg_depvar}; do
		if [ "$subpkg_dep" = "base" ]; then
			rdeps="$rdeps;$recipe_name"
		else
			rdeps="$rdeps;$recipe_name-$subpkg_dep"
		fi
	done
	
	split_dbg "Printing dependencies for '${package_name}_${package_vers}' (kind '$kind') sub-packages='$subpkgs'"
	if [ -z "$rdeps" ]; then
		split_dbg "        No dependencies to add, return."
		return
	else
		split_dbg "        Dependencies: $rdeps"
	fi
		
	for subpkg in $subpkgs; do
		split_dbg "Subpackage: $subpkg"
		if [ "$subpkg" = "base" ]; then
			subpkg=""
		else
			subpkg="-$subpkg"
		fi
		pkg=${package_name}${subpkg}_${package_vers}
		split_dbg "    Dependent package: $pkg"
		pkgsub="${pkgkind[${package_name}${subpkg}]}"
		if [ -z "${pkgsub}" ]; then
			split_dbg "        Skipping non-existent package '$pkg'"
			continue # Non-existent package... skip
		fi
		split_dbg "    Subpackage kind: $pkgsub"
		
		if [ "$pkgsub" = "base" -o "$pkgsub" = "shlibs" -o "$pkgsub" = "locale" ]; then 
			# shlibs and locale are base packages
			ekind=$kind
			pkgsub=""
		else
			if [[ " $needed_subpackages " =~ " $pkgsub " ]]; then
				ekind=$kind
			else
				if [ "$kind" = "rdepends" ]; then
					# Propagate dependencies as recommendations for non-needed subpackages
					ekind="rrecommends"
				else
					ekind=$kind
				fi
			fi
			pkgsub="-$pkgsub"
		fi
		split_dbg "    Rewritten subpackage kind: $pkgsub - $ekind"
	
		split_dbg "    Adding dependents to '$pkg.$ekind'"
		(
			local opened_line=false
			local in_vers=false
			for rdep in ${rdeps/;/ }; do
				case $rdep in
					\() echo -n " ( "
						in_vers=true
						;;
					\(*) echo -n " ( ${rdep#(}"
						in_vers=true
						;;
					\)) echo -n " )"
						in_vers=false
						;;
					*\)) echo -n " ${rdep%)} )"
						in_vers=false
						;;
					[a-zA-Z0-9]*)
						if $in_vers; then
							echo -n " $rdep"
							continue
						fi
						rdep_name=${rdep%%_*}
						rdep_vers=${rdep#*_}
						if [ "$rdep_vers" = "$rdep_name" ]; then
							if [[ " $rdep_name" =~ " $recipe_name-" ]]; then
								rdep_vers="_$package_vers"
							else
								rdep_vers=
							fi
						else
							rdep_vers="_$rdep_vers"
						fi
						split_dbg "        ${rdep_name}${pkgsub}${rdep_vers}"
						echo '' # Start a new line
						echo -n "${rdep_name}${pkgsub}${rdep_vers}" 
						;;
					*)	echo -n " $rdep"
						;;
				esac
			done
			echo ''
		) | append_dependencies "$metadataTempDir/$pkg.$ekind"
		split_dbg_cat "Generated '$pkg.$ekind':" $metadataTempDir/$pkg.$ekind
	done
}

empty=
split_package() {
	local package="$1"
	local pkg=${pkgkind[$package]}
	local base=${basepkg[$package]}
	local bin_package=${package}_${package_suffix}
	
	split_dbg "Generating '$bin_package' (type is '$pkg', base='$base')"
	
	local forced=true
	if [ "$pkg" = "base" -o "$pkg" = "shlibs" ]; then
		base_bin_packages="$base_bin_packages $bin_package"
	elif [[ " $needed_subpackages " =~ " $pkg " ]]; then
		forced_bin_packages="$forced_bin_packages $bin_package"
	elif [[ " $needed_subpackages " =~ " $package " ]]; then
		forced_bin_packages="$forced_bin_packages $bin_package"
	else
		other_bin_packages="$other_bin_packages $bin_package"
		forced=false
	fi

	# Create a file with the list of files to put in the archive
	# This has the additional benefit of allowing to create empty archives if needed
	local suffix
	local file_list=$contentsTempDir/$bin_package.list
	local pattern_name=pattern_${pkgname[$package]}
	split_dbg "Looking for pattern '$pattern_name'"
	local pattern_set="${!pattern_name+set}"
	if [ -z "$pattern_set" ]; then
		split_dbg "Pattern '$pattern_name' not set"
		local default_name
		if [ -z "$base" ]; then
			suffix="base"
			default_name=pattern_base
		elif [ "$base" = "$package_name" ]; then
			suffix=${pkgname[$package]#${pkgname[$base]}_}
			default_name=pattern_${suffix:-"base"}
		else
			# Here providing a default patter value is a bit
			# complex and would need detailed analysis of the
			# base package definition, but who really cares...
			log "Ignore unspecified package '$package' neither a base package nor a sub-package of '$package_name'"
			default_name="empty"
		fi
		split_dbg "	using '$default_name' instead"
		eval $pattern_name="'${!default_name}'"
	fi
	pattern_val="${!pattern_name}"

	# We add shlibs in the main pattern, so that non-splitted shlibs are always loaded
	# Note that the shlibs subpackage was always generated before the main package, so
	# its pattern was created by default above if not given in the recipe.
	local shlibs_pattern_name="${pattern_name}_shlibs"
	local shlibs_pattern_val="${!shlibs_pattern_name}"
	split_dbg "pattern_val='$pattern_val'"
	[ -n "$shlibs_pattern_val" ] && split_dbg "shlibs_pattern_val='$shlibs_pattern_val'"
	if [ -n "$pattern_val" -o -n "$shlibs_pattern_val" ]; then
		local pattern=
		# Transform paths to local paths in copied install directory...
		set -o noglob
		for fspec in $pattern_val $shlibs_pattern_val; do
			if [ "$fspec" = "/" ]; then
				pattern="${pattern# } ."
			else
				pattern="${pattern# } ${fspec#/}"
			fi
		done
		split_dbg "pattern='$pattern'"
		set +o noglob
		local fspec
		for fspec in $(echo ${pattern}); do
			if [ -d $fspec ]; then
				if [ "$fspec" != "${fspec%/}" ]; then
					# When specified with a trailing /, a directory must be defined in package, even if empty
					echo "${fspec%/}/$empty_dir_marker"
				fi
				find "$fspec" ! -type d
			elif [ -e "$fspec" -o -L "$fspec" ]; then
				echo "$fspec"
			fi
		done
	fi | sort | uniq > $file_list
	split_dbg_cat "File list in" "$file_list"
	
	if [ ! -s "$file_list" ] && ! $forced; then
		return 0 # Skip empty package
	fi

	# Generate the package description metadata and dependencies
	local spec
	for spec in SUMMARY DESCRIPTION SECTION HOMEPAGE LICENSE \
	            RPROVIDES RDEPENDS RRECOMMENDS RSUGGESTS \
	            RCONFLICTS RREPLACES
	do
		local spec_var=${spec}_${pkgname[$package]}
		local val_set=${!spec_var+set}
		if [ -z "$val_set" -a "$pkg" != "base" ]; then
			base=${basepkg[$package]}
			suffix=${pkgname[$package]#${pkgname[$base]}_}
			if [ -n "$base" ]; then
				case $spec in
					SUMMARY | DESCRIPTION | SECTION | HOMEPAGE | LICENSE)
						# Those can be inherited from their base package
						if [ "$suffix" != "${pkgname[$package]}" ]; then
							spec_var=${spec}_${pkgname[$package_name]}_${suffix}
							val_set=${!spec_var+set}
							if [ -z "$val_set" ]; then
								spec_var=${spec}_${suffix}
								val_set=${!spec_var+set}
							fi
						fi
						if [ -z "$val_set" ]; then
							spec_var=${spec}_${pkgname[$base]}
							val_set=${!spec_var+set}
						fi
						if [ -z "$val_set" ]; then
							spec_var=${spec}_${pkgname[$package_name]}
							val_set=${!spec_var+set}
						fi
						if [ -z "$val_set" ]; then
							spec_var=${spec}
							val_set=${!spec_var+set}
						fi
						;;
					*) # Others can only be inherited with suffix extension
						if [ "$suffix" != "${pkgname[$package]}" ]; then
							spec_var=${spec}_${pkgname[$base]}
							val_set=${!spec_var+set}
							if [ -n "$val_set" ]; then
								val_set="${!spec_var}"
								spec_var=${spec}_${pkgname[$package]}
								local subvar newvar=
								for subvar in $val_set; do
									newvar="$newvar $subvar-$suffix"
								done
								eval "${spec_var}='${newvar# }'"
								val_set=set
							fi
						fi
						;;
				esac
			fi
		fi
		case $spec in
			RDEPENDS | RRECOMMENDS | RSUGGESTS)
				# Sub-pkg dependencies may add more dependencies, so try to print them anyway
				print_dependencies $bin_package ${spec,,} "$needed_subpackages" ${!spec_var}
				;;
			RCONFLICTS | RREPLACES)
				if [ "$val_set" = "set" ]; then
					print_dependencies $bin_package ${spec,,} "$needed_subpackages" ${!spec_var}
				fi
				;;
			*)
				if [ "$val_set" = "set" ]; then
					lc_spec="${spec,,}"
					echo "${!spec_var}" > $metadataTempDir/$bin_package.$lc_spec
				fi
				;;
		esac
		split_dbg_cat "Current '${spec^^} metadata for '$bin_package':" $metadataTempDir/$bin_package.${spec,,}
	done

	# Don't forget to add the recipe runtime dependencies to the generated package
	print_dependencies $bin_package rdepends "base" ${recipe_runtime_dependencies}

	if [ "$pkg" = "shlibs" ]; then
		local shlibs
		# Here we must generate the splitted shlibs packages, 
		# as well as their dev, staticdev and dbg partners
		# Use a reversed sort to get longer names first
		shlibs=$(sed -n -e '/\.so$/s|^.*/\(lib[A-Za-z0-9+]\+\)[-_.].*\.so$|\1|p'				\
		                -e '/\.so$/s|^.*/\(lib[A-Za-z0-9+]\+\)\.so$|\1|p'						\
		                -e '/\.so[-.][0-9].*$/s|^.*/\(lib[A-Za-z0-9+]\+\).*\.so[-.].*$|\1|p'	\
		                $file_list | sort -r | uniq)
		split_dbg "shlibs='$shlibs'"
		
		local all_shlibs_pkgs=
		local shlib
		for shlib in ${shlibs}; do
			if [[ " ${unsplitted_shlibs} " =~ " ${shlib} " ]]; then
				split_dbg "Unsplitted shlib '$shlib'"
				continue # Keep these in the main package
			fi
			if [ " ${shlib} " = " ${package_name} " -o " ${shlib} " = " lib${package_name} " ]; then
				split_dbg "Main shlib '$shlib'"
				continue # This is the main package
			fi

			# Do not try to generate if an explicit package was defined
			local shlib_base="${package/shlibs/${shlib}}"
			if [ -n "${pkgkind[${shlib_base}]}" ]; then
				split_dbg "Package '$shlib_base' already defined"
				continue # Anyway this package should be empty (or the explicit package will be incomplete...)
			fi
		
			#
			# Generate list of all files for this shared library
			#
			local full_list=${file_list}.${shlib}-full
			grep -e "/${shlib}[-._]" $file_list | sort | uniq > "${full_list}"
			split_dbg_cat "Full shlib list in" "${full_list}"
		
			#
			# Shared objects 
			#
			# This includes all files matched by shlib except symbolic links in *.so
			#
			local so_list=${file_list/shlibs/${shlib}}
			grep -v "/${shlib}.so\$/" ${full_list} > ${so_list}
			local sofile
			for sofile in $(grep "/${shlib}.so\$/" ${full_list}); do
				if [ ! -L $sofile ]; then
					echo $sofile
				fi
			done >> ${so_list}
			if [ ! -s "${so_list}" ]; then
				rm -f ${full_list}
				split_dbg "No shared objects for '$shlib', skip entirely; files will go in the main shlibs package"
				continue # No shared objects, skip entirely; files will go in the main shlibs package
			fi
			split_dbg_cat "Shared objects list in" "${so_list}"
			# This is a new base package:
			shlib_pkg="${shlib_base}"
			subpkgs[${shlib_base}]="${subpkgs[${shlib_base}]} base"
			basepkg[${shlib_pkg}]="${package}"
			pkgkind[${shlib_pkg}]="shlibs"
			# Generate it
			bin_shlib_pkg=${shlib_pkg}_${package_suffix}
			split_dbg "bin_shlib_pkg=$bin_shlib_pkg"
			duplicate_metadata ${bin_package} ${bin_shlib_pkg} "shared objects for ${shlib}"
			build_package_tarball ${bin_shlib_pkg} ${so_list} ${shlib} base
			base_bin_packages="${base_bin_packages} ${bin_shlib_pkg}"
			# Eliminate packaged files from full list
			mv "$full_list" "$full_list.bak"
			comm -13 "${so_list}" "$full_list.bak" > "$full_list"
			rm -f "$full_list.bak"
			# Record it in the list of shared library packages
			all_shlibs_pkgs="$all_shlibs_pkgs $bin_shlib_pkg"
			
			#
			# Development files
			#
			local dev_list=${file_list/shlibs/${shlib}-dev}
			# Here I must also add the appropriate .so, .o and .la files, matching elements from the -dev subpkg,
			# as well as possible _nonshared.a files...
			# Currently we add all ".so" symlinks and .o or .la/_nonshared.a files having the shlib name as basename
			# but we should filter using the effective dev/staticdev patterns... 
			(
				grep "/${shlib}\.so\$/" ${full_list}
				grep "/${shlib}\.o\$/" ${full_list}
				grep "/${shlib}\.la\$/" ${full_list}
				grep "/${shlib}_nonshared\.a\$/" ${full_list}
				grep "/${shlib}-.*\.o\$/" ${full_list}
				grep "/${shlib}-.*\.la\$/" ${full_list}
				grep "/${shlib}_nonshared-.*\.a\$/" ${full_list}
				grep "/${shlib}\..*\.o\$/" ${full_list}
				grep "/${shlib}\..*\.la\$/" ${full_list}
				grep "/${shlib}_nonshared\..*\.a\$/" ${full_list}
				local dir
				for dir in lib usr/lib usr/*/lib; do
					if [ -d "$dir" ]; then 
						find $dir -name "${shlib}.so" -type l
						find $dir -name "${shlib}.o"
						find $dir -name "${shlib}.la"
						find $dir -name "${shlib}_nonshared.a"
						find $dir -name "${shlib}-*.o"
						find $dir -name "${shlib}-*.la"
						find $dir -name "${shlib}_nonshared-*.a"
						find $dir -name "${shlib}.*.o"
						find $dir -name "${shlib}.*.la"
						find $dir -name "${shlib}_nonshared.*.a"
					fi
				done
			) | sort | uniq > ${dev_list}
			split_dbg_cat "Development shlib links in" "${dev_list}"
			# This is a new derived package:
			shlib_pkg="${package/shlibs/${shlib}-dev}"
			subpkgs[${shlib_base}]="${subpkgs[${shlib_base}]} dev"
			basepkg[${shlib_pkg}]="${shlib_base}"
			pkgkind[${shlib_pkg}]="dev"
			# Generate it
			bin_shlib_pkg=${shlib_pkg}_${package_suffix}
			duplicate_metadata ${bin_package} ${bin_shlib_pkg} "development files for ${shlib}"
			build_package_tarball ${bin_shlib_pkg} ${dev_list} ${shlib} dev
			forced_bin_packages="${forced_bin_packages} ${bin_shlib_pkg}"
			# Eliminate packaged files from full list
			mv "$full_list" "$full_list.bak"
			comm -13 "${dev_list}" "$full_list.bak" > "$full_list"
			rm -f "$full_list.bak"
						
			#
			# Static libraries
			#
			local staticdev_list=${file_list/shlibs/${shlib}-staticdev}
			# Here I must also add the appropriate .a files, matching elements from the -dev/-staticdev subpkg
			# Currently we add all ".a" files having the shlib name as basename but we should filter using the
			# effective dev/staticdev patterns... 
			(
				grep "/${shlib}\.a\$/" ${full_list}
				grep "/${shlib}-.*\.a\$/" ${full_list}
				grep "/${shlib}\..*\.a\$/" ${full_list}
				local dir
				for dir in lib usr/lib usr/*/lib; do
					if [ -d "$dir" ]; then 
						find $dir -name "${shlib}.a"
						find $dir -name "${shlib}-*.a"
						find $dir -name "${shlib}.*.a"
					fi
				done
			) | sort | uniq > ${staticdev_list}
			split_dbg_cat "Static libraries in" "${staticdev_list}"
			# This is a new derived package:
			shlib_pkg="${package/shlibs/${shlib}-staticdev}"
			subpkgs[${shlib_base}]="${subpkgs[${shlib_base}]} staticdev"
			basepkg[${shlib_pkg}]="${shlib_base}"
			pkgkind[${shlib_pkg}]="staticdev"
			# Generate it
			bin_shlib_pkg=${shlib_pkg}_${package_suffix}
			duplicate_metadata ${bin_package} ${bin_shlib_pkg} "static library for ${shlib}"
			build_package_tarball ${bin_shlib_pkg} ${staticdev_list} ${shlib} staticdev
			forced_bin_packages="${forced_bin_packages} ${bin_shlib_pkg}"
			# Eliminate packaged files from full list
			mv "$full_list" "$full_list.bak"
			comm -13 "${staticdev_list}" "$full_list.bak" > "$full_list"
			rm -f "$full_list.bak"
			
			#
			# Debug info
			#
			local dbg_list=${file_list/shlibs/${shlib}-dbg}
			for sofile in $(cat ${so_list} ${dev_list}); do
				dbgfile=${sofile/\/lib${shlib}/.debug/lib${shlib}}
				if [ -r ${dbgfile} ]; then
					echo ${dbgfile}
				fi
			done | sort | uniq > ${dbg_list}
			split_dbg_cat "Debug shlib components in" "${dbg_list}"
			# This is a new derived package:
			shlib_pkg="${package/shlibs/${shlib}-dbg}"
			subpkgs[${shlib_base}]="${subpkgs[${shlib_base}]} dbg"
			basepkg[${shlib_pkg}]="${shlib_base}"
			pkgkind[${shlib_pkg}]="dbg"
			# Generate it
			bin_shlib_pkg=${shlib_pkg}_${package_suffix}
			duplicate_metadata $bin_package ${bin_shlib_pkg} "debug symbols for ${shlib}"
			build_package_tarball ${bin_shlib_pkg} ${dbg_list} ${shlib} dbg
			forced_bin_packages="${forced_bin_packages} ${bin_shlib_pkg}"
			
			# Now filter out the shlib from the global file
			cat ${so_list} ${dev_list} ${staticdev_list} ${dbg_list} | sort | uniq > ${full_list}
			sort ${file_list} | uniq > ${file_list}.sorted
			comm -13 ${full_list} ${file_list}.sorted > ${file_list}
			rm -rf ${file_list}.sorted ${full_list}
		done
		# Un splitted shared libraries go in the main package, not in the shlibs package
		# that is here only to allow loading ALL shared libraries at once
		rm -f ${file_list}
		touch ${file_list}
		
		#
		# Add a runtime dependency from the shlibs package to the splitted packages
		# so that loading xxxx-shlibs will load all shared libraries
		#
		split_dbg "Package '$bin_package' depends on '$all_shlibs_pkgs'"
		print_dependencies $bin_package rdepends "$needed_subpackages" $all_shlibs_pkgs
	fi

	# Here we must also intercept the locale sub-packages and split them further by language.
	# For all splitted packages, we must duplicate the metadata, but expanding the description 
	# by a line stating the contained library or the covered locale for locale we must keep
	# the non-langage-dependant files in the (forced) base locale package on which other depend
	if [ "$pkg" = "locale" ]; then
		# Here we must generate the splitted locale packages, 
		# as well as their dev and dbg partners
		local locales=$(sed -n -e 's|^.*/locale/\([A-Za-z0-9\@_-]\+\)/.*$|\1|p'	$file_list | sort | uniq)
		split_dbg "locales='$locales'"
		
		local all_locales_pkgs=
		local package_locales_file="$metadataTempDir/${package_name}_${package_suffix}.locales"
		touch "$package_locales_file"
		local locale
		for locale in ${locales}; do
			#
			# Generate list of all files for this locale
			#
			local full_list="${file_list/locale/locale-${locale//_/-}}"
			grep -e "/${locale}/" "$file_list" | sort | uniq > "${full_list}"
			split_dbg_cat "Full locale list in" "${full_list}"
		
			# This is a new locale package:
			local locale_pkg="${package/locale/locale-${locale//_/-}}"
			subpkgs[${locale_pkg}]="base"
			basepkg[${locale_pkg}]="${package}"
			pkgkind[${locale_pkg}]="locale"
			# Generate it
			local bin_locale_pkg=${locale_pkg}_${package_suffix}
			duplicate_metadata ${bin_package} ${bin_locale_pkg} "Language files for ${locale}"
			build_package_tarball ${bin_locale_pkg} "${full_list}" ${locale} base
			
			# Record it in the list of locale packages
			all_locales_pkgs="$all_locales_pkgs $bin_locale_pkg"
			
			# Now filter out the locales from the global file
			sort "${file_list}" | uniq > "${file_list}.sorted"
			comm -13 "${full_list}" "${file_list}.sorted" > "${file_list}"
			rm -rf "${file_list}.sorted"
			
			echo "$locale" >> "$package_locales_file"
		done
		
		# Unsplitted locale files go in the main package, not in the locale package
		# that is here only to allow loading ALL locales at once
		rm -f ${file_list}
		touch ${file_list}
		
		#
		# Add a runtime dependency from the locales package to the splitted package
		# so that loading xxxx-locale will load all locales
		#
		split_dbg "Package '$bin_package' depends on '$all_locales_pkgs'"
		print_dependencies $bin_package rdepends "$needed_subpackages" $all_locales_pkgs
	fi
	
	# Finally generate the package
	build_package_tarball $bin_package $file_list $package $pkg
	generate_preinst $bin_package $package
}

do_split() {
	generate_package_list
	add_missing_packages
	sort_packages
	for package in "${sorted_packages[@]}"; do
		log ""
		split_package $package
		log "Package=$package, Status=$?"
	done
	log ""
}

#
# Check that all files has been packaged
#
check_remaining_files() {
	local remaining_files=$(find . -type f | sed -e 's/^\.//')
	if [ -n "$remaining_files" ]; then
		echo "#################################################################################################"
		echo "#"
		echo "# There are unpackaged files in package ${package_name}"
		echo "#"
		echo "# Unpackaged files are:"
		local file
		for file in $remaining_files; do
			echo "# 	$file"
		done
		echo "#"
		echo "#################################################################################################"
		for package in "${sorted_packages[@]}"; do
			if [ "$package" = "$package_name" ]; then
				pattern_var="pattern_base"
			else
				pattern_var="pattern_$(shell_variable_name ${package})"
			fi
			if [ "${!pattern_var+set}}" = "set" ]; then
				log "Content of '$package' is '${!pattern_var}'"
			else
				log "Package '$package' automatically created"
			fi
		done
		fatal 1 "Can't split package '${package_name}'; please correct packaging directives"
	fi
}

# Ensure that file paths in contents files are absolute (they were created with relative paths)
adjust_content_files() {
	for cfile in $contentsTempDir/*.list; do
		mv $cfile $cfile.bak
		sed -e 's|^\([^/]\)|/\1|' $cfile.bak > $cfile
		rm -f $cfile.bak
	done
}

#
# Automatically create dynamic libraries runtime dependencies
#

# Warning: Here we can no more use installTempDir, as it was already
# emptied above, so look at the original installed files
get_needed_libraries() {
	local bin_package="$1"
	local forced="$2"
	local package base subpkg
	(
		for file in $(cat "$contentsTempDir/$bin_package.list"); do
		    call_readelf -d "${installDir%/}/${file#/}" 2> /dev/null    \
		    | grep -E '\(NEEDED\)' 2> /dev/null                         \
		    | sed -r -e 's/^.*:[[:space:]]+\[(.*)\]$/\1/;'
		done | sort | uniq | while read lib; do
			split_dbg "    Considering '$lib'"
			# Look for the package providing this library
			grep -l "/$lib\$" $contentsTempDir/*.list $allContentsDir/*/*.list 2> /dev/null
		done | sed -e 's|^.*/||'                \
		           -e '/-dbg_/d'                \
		           -e '/-dev_/d'                \
		           -e '/-staticdev_/d'          \
		           -e 's/.list$//'              \
		           -e '/^'"$bin_package"'$/d'   \
		           -e '/^[[:space:]]*$/d' 
		package="${bin_package%_${package_suffix}}"
		subpkg="${pkgkind[$package]}"
		base="${package%-$subpkg}"
		if [ "$subpkg" != "base" ]; then
			echo "${base}_${package_suffix}"
			if [ -n "$forced" ]; then
				if [[ " base shlibs " =~ " $subpkg " ]]; then
					true # It's a base package, although it should not happen we can safely ignore the error
				else
					if [ -z "$subpkg" -o -z "$base" ]; then
						internal_error "Error on '$bin_package': base='$base' - subpkg='$subpkg'"
					elif [ -s "$metadataTempDir/${base}_${package_suffix}.rdepends" ]; then
						cat "$metadataTempDir/${base}_${package_suffix}.rdepends" | sed -e "s/_/-${subpkg}_/"
					fi
				fi
			fi
		fi
	) | append_dependencies "$metadataTempDir/$bin_package.rdepends"
}

generate_automatic_runtime_dependencies() {
	log "Generating ${package_name} automatic runtime dependencies"
	for bin_package in $base_bin_packages $other_bin_packages; do
		get_needed_libraries $bin_package
	done
	for bin_package in $forced_bin_packages; do
		get_needed_libraries $bin_package forced
	done
}

install_split_package() {
	# Remove earlier definitions if any
	if [ -s "$metadataDir/$recipe_id.manifest" ]; then
		log_exec_locked $PLATFORM_LOC/feeds "(cd $packagesDir; rm -f \$(cat $metadataDir/$recipe_id.manifest))"
	fi
	log_exec_locked $PLATFORM_LOC/feeds rm -rf $contentsDir $tarballsDir $metadataDir
	
	# Now install the generated packages
	cd -
	log_exec rm -rf $installTempDir
	log_exec copy_tree_excluding -l $PLATFORM_LOC/feeds $tempDir $PLATFORM_LOC
	
	# And update the shared global list of locales
	if [ -s "$global_locales_file" ]; then
		merge_into_shared_list_file "$global_locales_file" "$package_locales_file"
	fi
}

do_split_append() {
	check_remaining_files
	adjust_content_files
	generate_automatic_runtime_dependencies
	install_split_package
}